#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for patching the configuration dictionary
"""
import os
from meerschaum.config._paths import PATCH_DIR_PATH, PERMANENT_PATCH_DIR_PATH
from meerschaum.config._read_config import read_config
patch_config = None
if PATCH_DIR_PATH.exists():
    from meerschaum.utils.yaml import yaml, _yaml
    if _yaml is not None:
        patch_config = read_config(directory=PATCH_DIR_PATH)

permanent_patch_config = None
if PERMANENT_PATCH_DIR_PATH.exists():
    from meerschaum.utils.yaml import yaml, _yaml
    if _yaml is not None:
        permanent_patch_config = read_config(directory=PERMANENT_PATCH_DIR_PATH)
else:
    permanent_patch_config = None


def apply_patch_to_config(
        config : dict,
        patch : dict
    ):
    """
    Patch the config dict with a new dict (cascade patching).
    """
    from meerschaum.utils.packages import attempt_import
    cascadict = attempt_import("cascadict", warn=True, install=True, color=False)
    base = cascadict.CascaDict(config)
    new = base.cascade(patch)
    return new.copy_flat()

def write_patch(
        patch : dict,
        debug : bool = False
    ):
    """
    Write patch dict to yaml
    """
    from meerschaum.config._edit import write_config
    if debug:
        from meerschaum.utils.formatting import pprint
        print(f"Writing configuration to {PATCH_DIR_PATH}:", file=sys.stderr)
        pprint(patch, stream=sys.stderr)
    write_config(patch, directory=PATCH_DIR_PATH)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Log into the API instance or refresh the token.
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Any

def login(
        self,
        debug : bool = False,
        **kw : Any
    ) -> SuccessTuple:
    """
    Log in and set the session token.
    """
    from meerschaum.utils.warnings import warn, info, error
    from meerschaum._internal.User import User
    from meerschaum.config.static import _static_config
    import json, datetime
    try:
        login_data = {
            'username' : self.username,
            'password' : self.password,
        }
    except AttributeError:
        return False, f"Please provide a username and password for '{self}' with `edit config`."
    response = self.post(
        _static_config()['api']['endpoints']['login'],
        data = login_data,
        use_token = False,
        debug = debug
    )
    if response:
        msg = f"Successfully logged into '{self}' as user '{login_data['username']}'."
        self._token = json.loads(response.text)['access_token']
        self._expires = datetime.datetime.strptime(
            json.loads(response.text)['expires'], 
            '%Y-%m-%dT%H:%M:%S.%f'
        )
    else:
        msg = (
            '' if self.get_user_id(
                User(self.username, self.password),
                use_token = False,
                debug = debug
            ) is not None else f"User '{self.username}' does not exist for '{self}'." + '\n'
        )
        msg += (
            f"   Failed to log into '{self}' as user '{login_data['username']}'.\n" +
            f"   Please verify login details for connector '{self}'."
        )
        warn(msg, stack=False)

    return response.__bool__(), msg

def refresh(
        self,
        debug : bool = False,
        **kw : Any,
    ):
    """
    Refresh the access token.
    """
    pass



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for copying elements.
"""

from __future__ import annotations
from meerschaum.utils.typing import Union, Any, Sequence, SuccessTuple, Optional, Tuple, List

def copy(
        action : Optional[List[str]] = None,
        **kw : Any
    ) -> SuccessTuple:
    """
    Copy pipes' attributes and make new pipes.

    Command:
        `copy {option}`

    Example:
        `copy pipes`
    """
    from meerschaum.utils.misc import choose_subaction
    if action is None:
        action = []
    options = {
        'pipes'      : _copy_pipes,
    }
    return choose_subaction(action, options, **kw)

def _copy_pipes(
        yes: bool = False,
        noask: bool = False,
        force: bool = False,
        debug: bool = False,
        **kw
    ) -> SuccessTuple:
    """
    Copy pipes' attributes and make new pipes.
    """
    from meerschaum import get_pipes, Pipe
    from meerschaum.utils.prompt import prompt, yes_no
    from meerschaum.utils.warnings import warn, info
    from meerschaum.utils.formatting import print_tuple
    from meerschaum.utils.formatting._shell import clear_screen
    pipes = get_pipes(as_list=True, **kw)
    successes = 0
    for p in pipes:
        ck = prompt(f"Connector keys for copy of pipe '{p}':", default=p.connector_keys)
        mk = prompt(f"Metric key for copy of pipe '{p}':", default=p.metric_key)
        lk = prompt(f"Location key for copy of pipe '{p}' (empty to omit):")
        if lk == '':
            lk = None
        _new_pipe = Pipe(
            ck, mk, lk,
            parameters=p.parameters.copy(),
        )
        instance_keys = prompt(f"Meerschaum instance to store new pipe '{_new_pipe}':", default=p.instance_keys)
        _new_pipe.instance_keys = instance_keys
        if _new_pipe.id is not None:
            warn(f"New pipe '{_new_pipe}' already exists. Skipping...", stack=False)
            continue
        _register_success_tuple = _new_pipe.register(debug=debug)
        if not _register_success_tuple[0]:
            warn(f"Failed to register new pipe '{_new_pipe}'.", stack=False)
            continue

        clear_screen(debug=debug)
        successes += 1
        print_tuple(
            (True, f"Successfully copied attributes of pipe '{p}' "
                + f"({p.instance_keys}) into '{_new_pipe}' ({_new_pipe.instance_keys}).")
        )
        if (
            force or yes_no(
                (
                    f"Do you want to copy data from pipe '{p}' into new pipe '{_new_pipe}'?\n\n"
                    + "If you specified `--begin`, `--end` or `--params`, data will be filtered."
                ),
                    noask=noask, yes=yes
                )
        ):
            _new_pipe.sync(p.get_data(debug=debug, **kw))

    msg = (
        "No pipes were copied." if successes == 0
        else (f"Copied {successes} pipe" + ("s" if successes != 1 else '') + '.')
    )

    return successes > 0, msg


### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.utils.misc import choices_docstring as _choices_docstring
copy.__doc__ += _choices_docstring('copy')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Synchronize a pipe's data with its source via its connector
"""

from __future__ import annotations

from meerschaum.utils.typing import (
    Union, Optional, Callable, Any, Tuple, SuccessTuple, Mapping, Dict, List
)

def sync(
        self,

        df : Optional[
            Union[
                pandas.DataFrame,
                Dict[str, List[Any]]
            ]
        ] = None,

        begin : Optional[
            datetime.datetime
        ] = None,

        end : Optional[
            datetime.datetime
        ] = None,

        force : bool = False,

        retries : int = 10,

        min_seconds : int = 1,

        check_existing : bool = True,

        blocking : bool = True,

        workers : Optional[int] = None,

        callback : Optional[
            Callable[[Tuple[bool, str]], Any]
        ] = None,

        error_callback : Optional[
            Callable[[Exception], Any]
        ] = None,

        chunksize : Optional[int] = -1,

        sync_chunks : bool = False,

        debug : bool = False,

        **kw : Any

    ) -> SuccessTuple:
    """
    Fetch new data from the source and update the pipe's table with new data.

    Get new remote data via fetch, get existing data in the same time period,
    and merge the two, only keeping the unseen data.

    :param df:
        An optional DataFrame to sync into the pipe. Defaults to None.

    :param begin:
        Optionally specify the earliest datetime to search for data.
        Defaults to None.

    :param end:
        Optionally specify the latest datetime to search for data.
        Defaults to None.

    :param force:
        If True, keep trying to sync untul `retries` attempts.
        Defaults to False.

    :param retries:
        If force, how many attempts to try syncing before declaring failure.
        Defaults to 10.

    :param min_seconds:
        If force, how many seconds to sleep between retries. Defaults to 1.

    :param check_existing:
        If True, pull and diff with existing data from the pipe.
        Defaults to True.

    :param blocking:
        If True, wait for sync to finish and return its result, otherwise
        asyncronously sync (oxymoron?) and return success. Defaults to True.

    :param workers:
        No use directly within `Pipe.sync()`. Instead is passed on to
        instance connectors' `sync_pipe()` methods (e.g. PluginConnector).
        Defaults to None.

    :param callback:
        Callback function which expects a SuccessTuple as input.
        Only applies when blocking = False.

    :param error_callback:
        Callback function which expects an Exception as input.
        Only applies when blocking = False.

    :param chunksize:
        Specify the number of rows to sync per chunk.
        If -1, resort to system configuration (default is 900).
        A `chunksize` of `None` will sync all rows in one transaction.
        Defaults to -1.

    :param sync_chunks:
        If possible, sync chunks in parallel.
        Defaults to False.

    :param debug:
        Verbosity toggle. Defaults to False.

    :param kw:
        Catch-all for keyword arguments.
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn, error
    import time
    if (callback is not None or error_callback is not None) and blocking:
        warn("Callback functions are only executed when blocking = False. Ignoring...")

    if (
          not self.connector_keys.startswith('plugin:')
          and not self.get_columns('datetime', error=False)
    ):
        return False, f"Cannot sync pipe '{self}' without a datetime column."

    ### add the stated arguments back into kw
    kw.update({
        'begin' : begin, 'end' : end, 'force' : force, 'retries' : retries,
        'min_seconds' : min_seconds, 'check_existing' : check_existing,
        'blocking' : blocking, 'workers' : workers, 'callback' : callback,
        'error_callback' : error_callback, 'sync_chunks' : (sync_chunks),
        'chunksize' : chunksize,
    })

    def _sync(
        p : 'meerschaum.Pipe',
        df : Optional['pandas.DataFrame'] = None
    ) -> SuccessTuple:
        ### ensure that Pipe is registered
        if not p.id:
            register_tuple = p.register(debug=debug)
            if not register_tuple[0]:
                return register_tuple

        ### If connector is a plugin with a `sync()` method, return that instead.
        ### If the plugin does not have a `sync()` method but does have a `fetch()` method,
        ### use that instead.
        ### NOTE: The DataFrame must be None for the plugin sync method to apply.
        ### If a DataFrame is provided, continue as expected.
        if df is None:
            try:
                if p.connector.type == 'plugin' and p.connector.sync is not None:
                    from meerschaum.utils.packages import activate_venv, deactivate_venv
                    activate_venv(p.connector.label, debug=debug)
                    return_tuple = p.connector.sync(p, debug=debug, **kw)
                    deactivate_venv(p.connector.label, debug=debug)
                    if not isinstance(return_tuple, tuple):
                        return_tuple = False, f"Plugin '{p.connector.label}' returned non-tuple value: {return_tuple}"
                    return return_tuple

            except Exception as e:
                msg = f"Failed to sync pipe '{p}' with exception: '" + str(e) + "'"
                if debug:
                    error(msg, silent=False)
                return False, msg

        ### default: fetch new data via the connector.
        ### If new data is provided, skip fetching
        if df is None:
            if p.connector is None:
                return False, f"Cannot fetch data for pipe '{p}' without a connector."
            df = p.fetch(debug=debug, **kw)
            if df is None:
                return False, f"Unable to fetch data for pipe '{p}'."
            if df is True:
                return True, f"Pipe '{p}' was synced in parallel."

        if debug:
            dprint(
                "DataFrame to sync:\n"
                + (str(df)[:255] + '...' if len(str(df)) >= 256 else str(df))
            )

        ### if force, continue to sync until success
        return_tuple = False, f"Did not sync pipe '{p}'."
        run = True
        _retries = 1
        while run:
            return_tuple = p.instance_connector.sync_pipe(
                pipe = p,
                df = df,
                debug = debug,
                **kw
            )
            _retries += 1
            run = (not return_tuple[0]) and force and _retries <= retries
            if run and debug:
                dprint(f"Syncing failed for pipe '{p}'. Attempt ( {_retries} / {retries} )")
                dprint(f"Sleeping for {min_seconds} seconds...")
                time.sleep(min_seconds)
            if _retries > retries:
                warn(
                    f"Unable to sync pipe '{p}' within {retries} attempt" +
                        ("s" if retries != 1 else "") + "!"
                )
        return return_tuple

    if blocking:
        return _sync(self, df = df)

    ### TODO implement concurrent syncing (split DataFrame? mimic the functionality of modin?)
    from meerschaum.utils.threading import Thread
    def default_callback(result_tuple : SuccessTuple):
        dprint(f"Asynchronous result from Pipe '{self}': {result_tuple}")
    def default_error_callback(x : Exception):
        dprint(f"Error received for Pipe '{self}': {x}")
    if callback is None and debug:
        callback = default_callback
    if error_callback is None and debug:
        error_callback = default_error_callback
    try:
        thread = Thread(
            target = _sync,
            args = (self,),
            kwargs = {'df' : df},
            daemon = False,
            callback = callback,
            error_callback = error_callback
        )
        thread.start()
    except Exception as e:
        return False, str(e)
    return True, f"Spawned asyncronous sync for pipe '{self}'."

def get_sync_time(
        self,
        params : Optional[Mapping[str, Any]] = None,
        debug : bool = False
    ) -> Optional['datetime.datetime']:
    """
    Get the most recent datetime value for a Pipe.

    :param params:
        Dictionary to build a WHERE clause for a specific column.
        E.g. params = None returns the latest possible datetime,
        but params = { 'a' : 1 } returns the latest datetime 'WHERE a = 1'.

    :param debug: Verbosity toggle.
    """
    from meerschaum.utils.warnings import error, warn
    if self.columns is None:
        warn(
            f"No columns found for Pipe '{self}'. " +
            "Pipe might not be registered or is missing columns in parameters."
        )
        return None

    if 'datetime' not in self.columns:
        warn(
            f"'datetime' must be declared in parameters:columns for Pipe '{self}'.\n\n" +
            f"You can add parameters for this Pipe with the following command:\n\n" +
            f"mrsm edit pipes -C {self.connector_keys} -M " +
            f"{self.metric_key} -L " +
            (f"[None]" if self.location_key is None else f"{self.location_key}")
        )
        return None

    return self.instance_connector.get_sync_time(
        self, params = params, debug = debug
    )

def exists(
        self,
        debug : bool = False
    ) -> bool:
    """
    See if a Pipe's table or view exists.
    """
    ### TODO test against views
    return self.instance_connector.pipe_exists(pipe=self, debug=debug)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Synchronize pipes' data with sources.

NOTE: `sync` required a SQL connection and is not intended for client use
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Any, List, Optional, Tuple

def sync(
        action : Optional[List[str]] = None,
        **kw : Any
    ) -> SuccessTuple:
    """
    Fetch and sync data for pipes.
    """
    from meerschaum.utils.misc import choose_subaction
    options = {
        'pipes'   : _sync_pipes,
    }
    return choose_subaction(action, options, **kw)

def _pipes_lap(
        workers : Optional[int] = None,
        debug : bool = None,
        unblock : bool = False,
        force : bool = False,
        min_seconds : int = 1,
        **kw : Any
    ) -> Tuple[List[meerschaum.Pipe], List[meerschaum.Pipe]]:
    """
    Do a lap of syncing pipes.
    """
    from meerschaum import get_pipes
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.packages import attempt_import
    from meerschaum.utils.formatting import print_tuple
    from meerschaum.utils.warnings import warn
    import time
    pipes = get_pipes(
        as_list = True,
        method = 'registered',
        debug = debug,
        **kw
    )

    def sync_pipe(p):
        """
        Wrapper function for the Pool.
        """
        from meerschaum.utils.warnings import warn
        try:
            return_tuple = p.sync(
                blocking = (not unblock),
                force = force,
                debug = debug,
                min_seconds = min_seconds,
                workers = workers,
                **kw
            )
        except Exception as e:
            import traceback
            traceback.print_exception(type(e), e, e.__traceback__)
            print("Error: " + str(e))
            return_tuple = (False, f"Failed to sync pipe '{p}' with exception:" + "\n" + str(e))

        print_tuple(return_tuple)
        return return_tuple

    from meerschaum.utils.pool import get_pool

    pool = get_pool(workers=workers)
    results = pool.map(sync_pipe, pipes) if pool is not None else [sync_pipe(p) for p in pipes]

    #  results = pool.map_async(sync_pipe, pipes)
    #  results = results.get()
    if results is None:
        warn(f"Failed to fetch results from syncing pipes.")
        succeeded_pipes = []
        failed_pipes = pipes
        results_dict = {
            p: (False, f"Could not fetch sync result for pipe '{p}'.") for p in pipes
        }
    else:
        ### Determine which pipes failed to sync.
        pipe_indices = [i for p, i in enumerate(pipes)]
        try:
            succeeded_pipes = [pipe_indices[i] for i, r in enumerate(results) if r[0]]
            failed_pipes = [pipe_indices[i] for i, r in enumerate(results) if not r[0]]
        except TypeError:
            succeeded_pipes = []
            failed_pipes = [p for p in pipes]
        results_dict = { p: r for p, r in zip(pipe_indices, results) }

    if len(failed_pipes) > 0:
        print("\n" + f"Failed to sync pipes:")
        for p in failed_pipes:
            print(f"  - {p}")

    if len(succeeded_pipes) > 0:
        success_msg = "\nSuccessfully synced pipes:"
        if unblock:
            success_msg = "\nSuccessfully spawned threads for pipes:"
        print(success_msg)
        for p in succeeded_pipes:
            print(f"  - {p}")

    if debug:
        from meerschaum.utils.formatting import pprint
        dprint("\n" + f"Return values for each pipe:")
        pprint(results_dict)

    return succeeded_pipes, failed_pipes

def _sync_pipes(
        loop : bool = False,
        min_seconds : int = 1,
        debug : bool = False,
        **kw : Any
    ) -> SuccessTuple:
    """
    Fetch and sync new data for pipes.
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn, info
    import time, sys
    run = True
    msg = ""
    interrupt_warning_msg = "Syncing was interrupted due to a keyboard interrupt."
    cooldown = 2 * (min_seconds + 1)
    success = []
    while run:
        lap_begin = time.time()
        try:
            success, fail = _pipes_lap(
                min_seconds = min_seconds,
                debug = debug,
                **kw
            )
        except Exception as e:
            import traceback
            traceback.print_exc()
            warn(
                f"Failed to sync all pipes. Waiting for {cooldown} seconds, then trying again.",
                stack = False
            )
            try:
                time.sleep(cooldown)
            except KeyboardInterrupt:
                warn(interrupt_warning_msg, stack=False)
                loop, run = False, False
            else:
                cooldown = int(cooldown * 1.5)
                continue
        except KeyboardInterrupt:
            warn(interrupt_warning_msg, stack=False)
            loop, run = False, False
            success, fail = None, None
        cooldown = 2 * (min_seconds + 1)
        lap_end = time.time()
        print()
        msg = (
            f"It took {round(lap_end - lap_begin, 2)} seconds to sync " +
            f"{len(success) + len(fail)} pipe" +
                ("s" if (len(success) + len(fail)) > 1 else "") + "\n" +
            f"    ({len(success)} succeeded, {len(fail)} failed)."
        ) if success is not None else "Syncing was aborted."
        if min_seconds > 0 and loop:
            print()
            info(
                f"Sleeping for {min_seconds} second" +
                ("s" if abs(min_seconds) != 1 else "")
                + '.'
            )
            try:
                time.sleep(min_seconds)
            except KeyboardInterrupt:
                loop, run = False, False
                warn(interrupt_warning_msg, stack=False)
        run = loop
    return (len(success) > 0 if success else False), msg

### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.utils.misc import choices_docstring as _choices_docstring
sync.__doc__ += _choices_docstring('sync')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
"""
Start the Meerschaum WebAPI with the `api` action.
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Optional, List, Any

def api(
        action : Optional[List[str]] = None,
        sysargs : Optional[List[str]] = None,
        debug : bool = False,
        mrsm_instance : Optional[str] = None,
        **kw : Any
    ) -> SuccessTuple:
    """
    Send commands to a Meerschaum WebAPI instance or boot a new instance

    Usage: `api [commands] {options}`
    Examples:
        - `api [start, boot, init]`
            - start the API server
        - `api show config`
            - execute `show config` on the `main` api instance
        - `api main show config`
            - see above
    
    If command is `start`, launch the Meerschaum WebAPI. If command is an api connector label,
        connect to that label. Otherwise connect to `main` api connector.
    """
    from meerschaum.utils.warnings import warn, info
    from meerschaum.utils.formatting import print_tuple
    if action is None:
        action = []
    if sysargs is None:
        sysargs = []
    if len(action) == 0:
        info(api.__doc__)
        return False, "Please provide a command to excecute (see above)"

    boot_keywords = {'start', 'boot', 'init'}
    if action[0] in boot_keywords:
        return _api_start(action=action, mrsm_instance=mrsm_instance, debug=debug, **kw)

    from meerschaum.config import get_config
    from meerschaum.connectors import get_connector
    import requests
    if debug:
        from meerschaum.utils.formatting import pprint
    api_configs = get_config('meerschaum', 'connectors', 'api', patch=True)

    api_label = "main"
    args_to_send = list(sysargs)
    ### remove `api`
    if 'api' in args_to_send:
        del args_to_send[0]
    if action[0] in api_configs:
        api_label = action[0]
        ### remove label from actions
        del action[0]
        if len(args_to_send) > 1:
            del args_to_send[0]
    kw['action'] = action
    kw['debug'] = debug
    kw['sysargs'] = args_to_send
    kw['yes'] = True
 
    api_conn = get_connector(type='api', label=api_label)
    
    if mrsm_instance is not None and str(mrsm_instance) == str(api_conn):
        warn(
            f"Cannot send Meerschaum instance keys '{mrsm_instance}' to itself. " +
            "Removing from arguments..."
        )
    elif mrsm_instance is not None:
        kw['mrsm_instance'] = str(mrsm_instance)

    success, message = api_conn.do_action(**kw)
    print_tuple((success, message), common_only=True)
    msg = f"Action " + ('succeeded' if success else 'failed') + " with message:\n" + str(message)
    return success, message

def _api_start(
        action : Optional[List[str]] = None,
        port : Optional[int] = None,
        workers : Optional[int] = None,
        mrsm_instance : Optional[str] = None,
        no_dash : bool = False,
        debug : bool = False,
        nopretty : bool = False,
        **kw : Any
    ) -> SuccessTuple:
    """
    Start the API server.

    Usage:
        `api start {options}`

    Options:
        - `-p, --port {number}`
            Port to bind the API server to.

        - `-w, --workers {number}`
            How many worker threads to run.
            Defaults to the number of CPU cores or 1 on Android.
    """
    from meerschaum.utils.packages import attempt_import, venv_contains_package, pip_install
    from meerschaum.utils.misc import is_int
    from meerschaum.utils.formatting import pprint, ANSI
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import error, warn
    from meerschaum.config import get_config, _config
    from meerschaum.config._paths import (
        API_UVICORN_CONFIG_PATH, CACHE_RESOURCES_PATH,
        PACKAGE_ROOT_PATH,
    )
    from meerschaum.config.static import _static_config
    from meerschaum.connectors.parse import parse_instance_keys
    from meerschaum.utils.pool import get_pool
    import os

    if action is None:
        action = []

    old_cwd = os.getcwd()
    if not debug:
        os.chdir(CACHE_RESOURCES_PATH)

    ### Uvicorn must be installed on the host because of multiprocessing reasons.
    uvicorn = attempt_import('uvicorn', venv=None, lazy=False)

    api_config = get_config('system', 'api')
    cf = _config()
    uvicorn_config = dict(api_config['uvicorn'])
    if port is None:
        ### default
        port = uvicorn_config['port']
        if len(action) > 1:
            if is_int(action[1]):
                port = int(action[1])

    pool = get_pool(workers=workers)
    if pool is None:
        workers = 1
    else:
        pool.close()
        pool.join()
    
    uvicorn_config['workers'] = workers

    uvicorn_config['debug'] = debug

    if mrsm_instance is None:
        mrsm_instance = get_config('meerschaum', 'api_instance', patch=True)

    ### Check if the API instance connector is another API
    instance_connector = parse_instance_keys(mrsm_instance, debug=debug)
    if instance_connector.type == 'api' and instance_connector.protocol != 'https':
        allow_http_parent = get_config(
            'system', 'api', 'permissions', 'chaining', 'insecure_parent_instance'
        )
        if not allow_http_parent:
            return False, (
                "Chaining Meerschaum API instances over HTTP is disabled!\n\n" +
                f"To use '{instance_connector}' as the Meerschaum instance for this API server, " +
                "please do one of the following:\n\n" +
                f"  - Ensure that '{instance_connector}' is available over HTTPS, " +
                "and with `edit config`,\n" +
                f"    change the `protocol` for '{instance_connector}' to 'https'.\n\n" +
                "  - Run `edit config system` and search for `permissions`.\n" +
                "    Under `api:permissions:chaining`, change the value of " +
                "`insecure_parent_instance` to `true`,\n" +
                "    then restart the API process."
            )

    uvicorn_config.update({
        'port' : port,
        'reload' : debug,
        'mrsm_instance' : mrsm_instance,
        'no_dash' : no_dash,
    })
    if debug:
        uvicorn_config['reload_dirs'] = [str(PACKAGE_ROOT_PATH)]
    uvicorn_config['use_colors'] = (not nopretty) if nopretty else ANSI

    api_config['uvicorn'] = uvicorn_config
    cf['system']['api']['uvicorn'] = uvicorn_config
    from meerschaum.api import __version__

    custom_keys = ['mrsm_instance', 'no_dash']

    ### write config to a temporary file to communicate with uvicorn threads
    import json, sys
    try:
        if API_UVICORN_CONFIG_PATH.exists() and not debug:
            os.remove(API_UVICORN_CONFIG_PATH)
            assert(not API_UVICORN_CONFIG_PATH.exists())
    except Exception as e:
        error(e)
    with open(API_UVICORN_CONFIG_PATH, 'w+') as f:
        if debug:
            dprint(f"Dumping API config file:", nopretty=nopretty)
            pprint(uvicorn_config, stream=sys.stderr, nopretty=nopretty)
        json.dump(uvicorn_config, f)

    ### remove custom keys before calling uvicorn
    for k in custom_keys:
        del uvicorn_config[k]

    if debug:
        dprint(f"Connecting to Meerschaum instance: '{mrsm_instance}'.", nopretty=nopretty)
        dprint(
            f"Starting Meerschaum API v{__version__} with the following configuration:",
            nopretty = nopretty
        )
        pprint(uvicorn_config, stream=sys.stderr, nopretty=nopretty)

    try:
        uvicorn.run(**uvicorn_config)
    except KeyboardInterrupt:
        pass

    os.chdir(old_cwd)

    return (True, "Success")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Drop a Pipe's table but keep its registration
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Any

def drop(
        self,
        debug : bool = False,
        **kw : Any
    ) -> SuccessTuple:
    """
    Call the Pipe's instance connector's drop_pipe method
    """
    return self.instance_connector.drop_pipe(self, debug=debug, **kw)



#! /usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
mrsm CLI entrypoint.
"""

import sys, os

def main() -> None:
    """
    Main CLI entry point.
    """
    old_cwd = os.getcwd()
    sysargs = sys.argv[1:]

    ### Check for a custom configuration directory.
    if '--root-dir' in sysargs:
        from meerschaum.actions.arguments._parse_arguments import parse_arguments
        from meerschaum.config._paths import set_root
        import pathlib
        from meerschaum.config.static import _static_config
        env_var = _static_config()['environment']['root']
        if env_var in os.environ:
            print(f"WARNING: '{env_var}' is set, so --root-dir will be ignored.")
        else:
            args = parse_arguments(sysargs)
            config_dir_path = pathlib.Path(args['root_dir']).absolute()
            if not config_dir_path.exists():
                print(
                    f"Invalid config directory '{str(config_dir_path)}'.\n" +
                    "Please enter a valid path for `--config-dir`."
                )
                sys.exit(1)
            set_root(config_dir_path)

    ### Catch help flags.
    if '--help' in sysargs or '-h' in sysargs:
        from meerschaum.actions.arguments._parser import parse_help
        parse_help(sysargs)
        return _exit(old_cwd=old_cwd)

    ### Catch version flags.
    if '--version' in sysargs or '-V' in sysargs:
        from meerschaum.actions.arguments._parser import parse_version
        parse_version(sysargs)
        return _exit(old_cwd=old_cwd)

    if ('-d' in sysargs or '--daemon' in sysargs) and ('stack' not in sysargs):
        from meerschaum.utils.daemon import daemon_entry
        daemon_entry(sysargs)
        return _exit(old_cwd=old_cwd)

    from meerschaum.actions import entry, get_shell

    ### Try to launch a shell if --shell is provided.
    if len(sysargs) == 0 or '--shell' in sysargs:
        if '--shell' in sysargs:
            sysargs.remove('--shell')
        get_shell(sysargs).cmdloop()
        return _exit(old_cwd=old_cwd)

    ### Print success or failure message.
    return_tuple = entry(sysargs)
    rc = 0
    if isinstance(return_tuple, tuple):
        from meerschaum.utils.formatting import print_tuple
        print_tuple(return_tuple, upper_padding=1)
        rc = 0 if (return_tuple[0] is True) else 1

    return _exit(rc, old_cwd=old_cwd)

def _exit(return_code : int = 0, old_cwd : str = None) -> None:
    _close_pools()
    ### Restore the previous working directory.
    if old_cwd is not None and old_cwd != os.getcwd():
        os.chdir(old_cwd)
    sys.exit(return_code)

def _close_pools():
    """
    Close multiprocessing pools before exiting.
    """
    ### Final step: close global pools.
    from meerschaum.utils.pool import get_pools
    for class_name, pool in get_pools().items():
        try:
            pool.close()
            pool.join()
        except Exception as e:
            pass

if __name__ == "__main__":
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define SQLAlchemy tables
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, Dict, Union

### store a tables dict for each connector
connector_tables = dict()

_sequence_flavors = {'duckdb'}

def get_tables(
        mrsm_instance : Optional[Union[str, meerschaum.connectors.Connector]] = None,
        create : bool = True,
        debug : Optional[bool] = None
    ) -> Dict[str, sqlalchemy.Table]:
    """
    Substantiate and create sqlalchemy tables
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.formatting import pprint
    from meerschaum.utils.warnings import warn, error
    from meerschaum.connectors.parse import parse_instance_keys
    from meerschaum.utils.packages import attempt_import
    from meerschaum import get_connector

    sqlalchemy, sqlalchemy_dialects_postgresql, sqlalchemy_utils = attempt_import(
        'sqlalchemy',
        'sqlalchemy.dialects.postgresql',
        'sqlalchemy_utils',
    )
    if not sqlalchemy:
        error(f"Failed to import sqlalchemy. Is sqlalchemy installed?")

    if mrsm_instance is None:
        conn = get_connector(debug=debug)
    elif isinstance(mrsm_instance, str):
        conn = parse_instance_keys(mrsm_instance, debug=debug)
    else: ### NOTE: mrsm_instance MUST BE a SQL Connector for this to work!
        conn = mrsm_instance

    ### kind of a hack. Create the tables remotely
    from meerschaum.connectors.api import APIConnector
    if isinstance(conn, APIConnector):
        return conn.create_metadata(debug=debug)

    global connector_tables
    if conn not in connector_tables:
        if debug:
            dprint(f"Creating tables for connector '{conn}'.")

        id_type = sqlalchemy.Integer
        params_type = (
            sqlalchemy_utils.types.json.JSONType if conn.flavor not in ('duckdb')
            else sqlalchemy.String
        )

        id_names = ('user_id', 'plugin_id', 'pipe_id')
        sequences = {
            k: sqlalchemy.Sequence(k + '_seq')
                for k in id_names 
        }
        id_col_args = { k: [k, id_type] for k in id_names }
        id_col_kw = { k: {'primary_key': True} for k in id_names }

        if conn.flavor in _sequence_flavors:
            for k, args in id_col_args.items():
                args.append(sequences[k])
            for k, kw in id_col_kw.items():
                kw.update({'server_default': sequences[k].next_value()})

        _tables = {
            'users' : sqlalchemy.Table(
                'users',
                conn.metadata,
                sqlalchemy.Column(
                    *id_col_args['user_id'],
                    **id_col_kw['user_id'],
                ),
                sqlalchemy.Column('username', sqlalchemy.String, index=True, nullable=False),
                sqlalchemy.Column('password_hash', sqlalchemy.String),
                sqlalchemy.Column('email', sqlalchemy.String),
                sqlalchemy.Column('user_type', sqlalchemy.String),
                sqlalchemy.Column('attributes', params_type),
                extend_existing = True,
            ),
            'plugins' : sqlalchemy.Table(
                'plugins',
                conn.metadata,
                sqlalchemy.Column(
                    *id_col_args['plugin_id'],
                    **id_col_kw['plugin_id'],
                ),
                sqlalchemy.Column('plugin_name', sqlalchemy.String, index=True, nullable=False),
                sqlalchemy.Column('user_id', sqlalchemy.Integer, nullable=False),
                sqlalchemy.Column('version', sqlalchemy.String),
                sqlalchemy.Column('attributes', params_type),
                extend_existing = True,
            ),
        }

        _tables['pipes'] = sqlalchemy.Table(
            "pipes",
            conn.metadata,
            sqlalchemy.Column(
                *id_col_args['pipe_id'],
                **id_col_kw['pipe_id'],
            ),
            sqlalchemy.Column("connector_keys", sqlalchemy.String, index=True, nullable=False),
            sqlalchemy.Column("metric_key", sqlalchemy.String, index=True, nullable=False),
            sqlalchemy.Column("location_key", sqlalchemy.String, index=True),
            sqlalchemy.Column("parameters", params_type),
            #  sqlalchemy.UniqueConstraint(
                #  'connector_keys', 'metric_key', 'location_key', name='pipe_index'
            #  ),
            extend_existing = True,
        )

        try:
            conn.metadata.create_all(bind=conn.engine)
        except Exception as e:
            #  import traceback
            #  traceback.print_exception(type(e), e, e.__traceback__)
            warn(str(e))

        ### store the table dict for reuse (per connector)
        connector_tables[conn] = _tables

    return connector_tables[conn]



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Generic Connector parent class. Reads attributes from the configuration.
Consult implemented child classes like SQLConnector, APIConnector, MQTTConnector, or
PluginConnector for further details.
"""

from __future__ import annotations
import abc
from meerschaum.utils.typing import Iterable, Optional, Any, Union

class Connector(metaclass=abc.ABCMeta):
    def __init__(
            self,
            type : Optional[str] = None,
            label : Optional[str] = None,
            **kw : Any
        ):
        """
        type : str
            The type of the connection. Used as a key in config.yaml to get attributes.
            Supported values are 'sql', 'api', 'mqtt', 'plugin'.

        label : str
            The label for the connection. Used as a key within config.yaml

        pandas : str
            Custom pandas implementation name.
            May change to modin.pandas soon.

        If config.yaml is set for the given type and label, the hierarchy looks like so:
        meerschaum:
            connections:
                {type}:
                    {label}:
                        ### attributes go here

        Read config.yaml for attributes partitioned by connection type and connection label.
        Example: type="sql", label="main"
        """
        self._original_dict = self.__dict__.copy()
        self._set_attributes(type=type, label=label, **kw)

    def _reset_attributes(self):
        self.__dict__ = self._original_dict

    def _set_attributes(
            self,
            type : Optional[str] = None,
            label : str = "main",
            pandas : Optional[str] = None,
            inherit_default : bool = True,
            **kw : Any
        ):
        from meerschaum.utils.warnings import error
        if label == 'default':
            error("Label cannot be 'default'. Did you mean 'main'?")
        self.type, self.label = type, label

        from meerschaum.config import get_config
        conn_configs = get_config('meerschaum', 'connectors')
        connector_config = get_config('system', 'connectors')

        ### inherit attributes from 'default' if exists
        if inherit_default:
            inherit_from = 'default'
            if self.type in conn_configs and inherit_from in conn_configs[self.type]:
                _inherit_dict = conn_configs[self.type][inherit_from].copy()
                self.__dict__.update(_inherit_dict)

        ### load user config into self.__dict__
        if self.type in conn_configs and self.label in conn_configs[self.type]:
            self.__dict__.update(conn_configs[self.type][self.label])

        ### load system config into self.sys_config
        ### (deep copy so future Connectors don't inherit changes)
        if self.type in connector_config:
            from copy import deepcopy
            self.sys_config = deepcopy(connector_config[self.type])

        ### add additional arguments or override configuration
        self.__dict__.update(kw)

    def verify_attributes(
            self,
            required_attributes : set = {
                'label'
            },
            debug : bool = False
        ):
        """
        Ensure that the required attributes have been met.
        
        required_attributes : set
            Attributes to be verified.

        The Connector base class checks the minimum requirements.
        Child classes may enforce additional requirements.
        """
        from meerschaum.utils.warnings import error, warn
        from meerschaum.utils.debug import dprint
        missing_attributes = set()
        for a in required_attributes:
            if a not in self.__dict__:
                missing_attributes.add(a)
        if len(missing_attributes) > 0:
            error(
                f"Please provide connection configuration for connector '{self.type}:{self.label}' "
                f"in the configuration file (open with `mrsm edit config`) or as arguments for the Connector.\n\n"
                f"Missing attributes: {missing_attributes}",
                silent = True,
                stack = False
            )

    def fetch(
            self,
            instructions : dict
        ) -> 'pd.DataFrame':
        """
        Abstract method for all Connectors.
        In addition to specialized functionality, all Connectors must be able to fetch data
        based on criteria provided in the instructions dictionary.

        Returns pandas (or pandas derivative) DataFrame
        """
        from meerschaum.utils.warnings import error
        error("fetch() must be implemented in children classes", NotImplementedError)

    def __str__(self):
        return f"{self.type}:{self.label}"

    def __repr__(self):
        return str(self)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
<img src="https://meerschaum.io/assets/banner_1920x320.png" alt="Meerschaum banner">

| PyPI                                                         | Repository                                                   | License                                                      |
| ------------------------------------------------------------ | ------------------------------------------------------------ | ------------------------------------------------------------ |
| ![PyPI]( https://img.shields.io/pypi/v/meerschaum?color=%2300cc66&label=Version ) | ![GitHub repo size](https://img.shields.io/github/repo-size/bmeares/Meerschaum?label=Repo%20size) | ![PyPI - License](https://img.shields.io/pypi/l/meerschaum?label=License) |
| ![PyPI - Python Version]( https://img.shields.io/pypi/pyversions/meerschaum?label=Python&logo=python&logoColor=%23ffffff ) | ![Lines of code]( https://img.shields.io/tokei/lines/github/bmeares/Meerschaum?label=Total%20Lines ) |                                                              |


# What is Meerschaum?
Meerschaum is a platform for quickly creating and managing time-series data streams called **pipes**. With Meerschaum, you can have a data visualization stack running in minutes.

The power of the Meerschaum system comes from projects like [pandas](https://pandas.pydata.org/), [sqlalchemy](https://www.sqlalchemy.org/), [fastapi](https://fastapi.tiangolo.com/), and more.

# Why Meerschaum?

If you've worked with time-series data, you know the headaches that come with ETL. Meerschaum is a system that makes consolidating and syncing data easy.

Anytime you need to move time-series data from somewhere like a database, API, MQTT topic, or even CSV files (we've all done it at some point), you usually end up writing some scripts that silently break in a year. Meerschaum instead gives you better tools to define and sync your data streams. And don't worry — you can always incorporate Meerschaum into your existing scripts.

# Quick Start

For a more thorough setup guide, visit the [Getting Started](https://meerschaum.io/get-started/) page at [meerschaum.io](https://meerschaum.io).

## TL;DR

```bash
pip install -U --user meerschaum
mrsm stack up -d
mrsm bootstrap pipes
```

## Usage Documentation

Please visit [meerschaum.io](https://meerschaum.io) for setup, usage, and troubleshooting information. You can find technical documentation at [docs.meerschaum.io](https://docs.meerschaum.io).

## Plugins

Meerschaum has a straightforward plugin system. For details on installing, using, and writing plugins, check out the [plugins documentation](https://meerschaum.io/reference/plugins/types-of-plugins) at [meerschaum.io](https://meerschaum.io).

# Support Meerschaum's Development

I'm a full-time graduate student, and I work on Meerschaum in my free time. If you enjoy Meerschaum and want to support its development, you can [buy me a beer (or coffee)](https://www.buymeacoffee.com/bmeares).
"""



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
This package includes argument parsing utilities
"""

from meerschaum.actions.arguments._parse_arguments import parse_arguments, parse_line
from meerschaum.actions.arguments._parser import parser, add_plugin_argument



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for interacting with pipes via the web interface.
"""

from __future__ import annotations
import shlex
from dash.dependencies import Input, Output, State
from meerschaum.utils.typing import List, Optional, Dict, Any, Tuple, Union
from meerschaum.utils.misc import string_to_dict
from meerschaum.utils.packages import attempt_import
from meerschaum.api.dash import (
    dash_app, debug, _get_pipes
)
from meerschaum.api.dash.connectors import get_web_connector
from meerschaum.api.dash.components import alert_from_success_tuple
dbc = attempt_import('dash_bootstrap_components', lazy=False)

def keys_from_state(
        state : Dict[str, Any],
        with_params : bool = False
    ) -> Union[
        Tuple[List[str], List[str], List[str]],
        Tuple[List[str], List[str], List[str], str],
    ]:
    """
    Read the current state and return the selected keys lists.
    """
    _filters = {
        'ck' : state[f"connector-keys-{state['pipes-filter-tabs.active_tab']}.value"],
        'mk' : state[f"metric-keys-{state['pipes-filter-tabs.active_tab']}.value"],
        'lk' : state[f"location-keys-{state['pipes-filter-tabs.active_tab']}.value"],
    }
    if state['pipes-filter-tabs.active_tab'] == 'input':
        try:
            #  params = string_to_dict(state['params-textarea.value'])
            params = string_to_dict(state['search-parameters-editor.value'])
        except Exception as e:
            params = None
    else:
        params = None

    for k in _filters:
        _filters[k] = [] if _filters[k] is None else _filters[k]
        if not isinstance(_filters[k], list):
            try:
                _filters[k] = shlex.split(_filters[k])
            except Exception as e:
                print(e)
                _filters[k] = []
    keys = [_filters['ck'], _filters['mk'], _filters['lk']]
    if with_params:
        keys.append(params)
    return tuple(keys)

def pipes_from_state(
        state : Dict[str, Any],
        **kw
    ):
    """
    Return a pipes dictionary or list from get_pipes.
    """
    _ck, _mk, _lk, _params = keys_from_state(state, with_params=True)
    try:
        _pipes = _get_pipes(
            _ck, _mk, _lk,
            params = _params,
            mrsm_instance = get_web_connector(state), 
            **kw
        )
    except Exception as e:
        return False, str(e)
    return _pipes

def get_pipes_cards(*keys):
    """
    Returns a tuple:
        - pipes as a list of cards
        - alert list
    """
    _pipes = pipes_from_state(*keys, as_list=True)
    alert = alert_from_success_tuple(_pipes)
    if not isinstance(_pipes, list):
        _pipes = []
    return ([dbc.Card(str(p)) for p in _pipes], [alert])




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Attempt to create a pipe's requirements in one method.
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Optional

def bootstrap(
        self,
        debug : bool = False,
        yes : bool = False,
        force : bool = False,
        noask : bool = False,
        **kw
    ) -> SuccessTuple:
    """
    Prompt the user to create a pipe's requirements all from one method.
    This method shouldn't be used in any automated scripts because it interactively
    prompts the user and therefore may hang.
    """

    from meerschaum.utils.warnings import warn, info, error
    from meerschaum.utils.prompt import prompt, yes_no
    from meerschaum.utils.formatting import pprint
    from meerschaum.config import get_config
    from meerschaum.utils.formatting._shell import clear_screen
    from meerschaum.utils.formatting import print_tuple

    _clear = get_config('shell', 'clear_screen', patch=True)

    if self.get_id(debug=debug) is not None:
        delete_tuple = self.delete(debug=debug)
        if not delete_tuple[0]:
            return delete_tuple

    info(f"Please enter the following column names for pipe '{self}':")
    while True:
        try:
            datetime_name = prompt(f"Datetime:", icon=False)
        except KeyBoardError:
            return False, "Failed to bootstrap pipe '{self}':\n" + str(e)
        if datetime_name == '':
            warn(f"Please enter a datetime column.", stack=False)
            continue

        try:
            id_name = prompt(f"ID (empty to omit):", icon=False)
        except KeyBoardError:
            return False, f"Failed to bootstrap pipe '{self}':\n" + str(e)
        if id_name == '':
            id_name = None
        break

    if _clear:
        clear_screen(debug=debug)

    _parameters = {
        'columns' : {
            'datetime' : datetime_name,
            'id' : id_name,
        },
    }

    parameters_update_types = {
        'sql'  : {
            'fetch' : {
                'definition' : None,
            },
        },
        'api'  : {
            'fetch' : {
                'connector_keys' : None,
                'metric_key' : None,
                'location_key' : None,
            },
        },
        'mqtt' : {
            'fetch' : {
                'topic' : '#',
            },
        },
    }

    _parameters.update(parameters_update_types.get(self.connector.type, {}))
    self.parameters = _parameters
    pprint(self.parameters)
    prompt(f"\nPress [Enter] to register pipe '{self}' with the above configuration:", icon=False)
    register_tuple = self.register(debug=debug)
    if not register_tuple[0]:
        return register_tuple

    if _clear:
        clear_screen(debug=debug)

    try:
        if yes_no(
            f"Would you like to edit the definition for pipe '{self}'?", yes=yes, noask=noask
        ):
            edit_tuple = self.edit_definition(debug=debug)
            if not edit_tuple[0]:
                return edit_tuple

            if yes_no(f"Would you like to sync pipe '{self}' now?", yes=yes, noask=noask):
                sync_tuple = self.sync(debug=debug)
                if not sync_tuple[0]:
                    return sync_tuple
    except Exception as e:
        return False, f"Failed to bootstrap pipe '{self}':\n" + str(e)

    print_tuple((True, f"Finished bootstrapping pipe '{self}'!"))
    info(
        f"You can edit this pipe later with `edit pipes` or set the definition with `edit pipes definition`.\n" +
        "To sync data into your pipe, run `sync pipes`."
    )

    return True, "Success"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
The entry point for launching Meerschaum actions.
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, List, Optional

def _entry(sysargs : Optional[List[str]] = None) -> SuccessTuple:
    """
    Parse arguments and launch a Meerschaum action.
    The `action` list removes the first element.

    Examples of action:
        'show actions' -> ['actions']
        'show' -> []
    """
    from meerschaum.actions.arguments import parse_arguments
    if sysargs is None:
        sysargs = []
    if not isinstance(sysargs, list):
        import shlex
        sysargs = shlex.split(sysargs)
    args = parse_arguments(sysargs)
    return _entry_with_args(**args)

def _entry_with_args(**kw) -> SuccessTuple:
    """
    Execute a Meerschaum action with keyword arguments.
    Use `_entry()` for parsing sysargs before executing.
    """
    from meerschaum.actions import actions, original_actions, get_shell
    from meerschaum.utils.packages import activate_venv, deactivate_venv
    import sys
    if kw.get('trace', None):
        from meerschaum.utils.misc import debug_trace
        debug_trace()
    if len(kw.get('action', [])) == 0:
        return get_shell().cmdloop()
    #  if kw.get('daemon', False):
        #  from meerschaum.utils.daemon import daemon_action
        #  return daemon_action(**kw)

    main_action = kw['action'][0]

    ### if action does not exist, execute in bash
    if main_action not in actions:
        main_action = 'sh'
        kw['action'].insert(0, main_action)

    ### Check if the action is a plugin, and if so, activate virtual environment.
    plugin_name = (
        actions[main_action].__module__.split('.')[-1] if (
            actions[main_action].__module__.startswith('plugins.')
        ) else None
    )

    del kw['action'][0]

    activate_venv(venv=plugin_name, debug=kw.get('debug', False))

    try:
        result = actions[main_action](**kw)
    except Exception as e:
        if kw.get('debug', False):
            import traceback
            traceback.print_exception(type(e), e, e.__traceback__)
        result = False, (
            f"Failed to execute '{' '.join([main_action] + kw['action'])}' with exception:\n\n" +
            f"'{e}'.\n\nRun again with '--debug' to see a full stacktrace."
        )

    deactivate_venv(venv=plugin_name, debug=kw.get('debug', False))

    return result




#! /usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
This module contains functions for parsing arguments
"""

from __future__ import annotations
from meerschaum.utils.typing import List, Dict, Any, Optional

from meerschaum.actions.arguments._parser import parser

def parse_arguments(sysargs : List[str]) -> dict[str, Any]:
    """
    Parse a list of arguments into standard Meerschaum arguments.
    Returns a dictionary of argument_name -> argument_value.

    :param sysargs:
        List of command-line arguments to process. Does not include the executable.
        E.g. ['show', 'version', '--nopretty']
    """
    import copy
    from meerschaum.config.static import _static_config

    sub_arguments = []
    sub_arg_indices = []
    begin_decorator, end_decorator = _static_config()['system']['arguments']['sub_decorators']
    found_begin_decorator = False
    for i, word in enumerate(sysargs):
        is_sub_arg = False
        if not found_begin_decorator:
            found_begin_decorator = word.startswith(begin_decorator)
            found_end_decorator = word.endswith(end_decorator)

        if found_begin_decorator:
            ### check if sub arg is ever closed
            for a in sysargs[i:]:
                if a.endswith(end_decorator):
                    is_sub_arg = True
                    found_begin_decorator = False
        elif found_end_decorator:
            for a in sysargs[:i]:
                if a.startswith(begin_decorator):
                    is_sub_arg = True
                    found_begin_decorator = False
        if is_sub_arg:
            ### remove decorators
            sa = word
            if sa.startswith(begin_decorator):
                sa = sa[len(begin_decorator):]
            if sa.endswith(end_decorator):
                sa = sa[:-1 * len(end_decorator)]
            sub_arguments.append(sa)
            ### remove sub-argument from action list
            sub_arg_indices.append(i)

    ### rebuild sysargs without sub_arguments
    filtered_sysargs = [
        word for i, word in enumerate(sysargs)
            if i not in sub_arg_indices
    ]

    args, unknown = parser.parse_known_args(filtered_sysargs)

    ### if --config is not empty, cascade down config
    ### and update new values on existing keys / add new keys/values
    if args.config is not None:
        from meerschaum.config._patch import write_patch, apply_patch_to_config
        from meerschaum.config._paths import PATCH_DIR_PATH
        from meerschaum.utils.packages import reload_package
        import os, meerschaum.config, shutil
        write_patch(args.config)
        reload_package('meerschaum')
        ### clean up patch so it's not loaded next time
        if PATCH_DIR_PATH.exists():
            shutil.rmtree(PATCH_DIR_PATH)

    args_dict = vars(args)
    args_dict['sysargs'] = sysargs
    args_dict['filtered_sysargs'] = filtered_sysargs
    ### append decorated arguments to sub_arguments list
    if 'sub_args' not in args_dict:
        args_dict['sub_args'] = []
    if args_dict['sub_args'] is None:
        args_dict['sub_args'] = []
    sub_arguments = args_dict['sub_args'] + sub_arguments
    parsed_sub_arguments = []
    for sub_arg in sub_arguments:
        if ' ' in sub_arg:
            parsed_sub_arguments += sub_arg.split(' ')
        else:
            parsed_sub_arguments.append(sub_arg)
    args_dict['sub_args'] = parsed_sub_arguments
    ### In case of empty subargs
    if args_dict['sub_args'] == ['']:
        args_dict['sub_args'] = []

    ### remove None (but not False) args
    none_args = []
    for a, v in args_dict.items():
        if v is None:
            none_args.append(a)
    for a in none_args:
        del args_dict[a]

    ### location_key '[None]' or 'None' -> None
    if 'location_keys' in args_dict:
        args_dict['location_keys'] = [
            None if lk in ('[None]', 'None')
            else lk for lk in args_dict['location_keys'] 
        ]

    return parse_synonyms(args_dict)

def parse_line(line : str) -> dict:
    """
    Parse a line of text into standard Meerschaum arguments.

    line: str
        Line of text to be parsed

    returns: dict of arguments
    """
    import shlex
    try:
        args_list = shlex.split(line)
        return parse_arguments(
            shlex.split(line)
        )
    except Exception as e:
        return {'action' : [], 'text' : line,}

def parse_synonyms(
        args_dict : Dict[str, Any]
    ) -> Dict[str, Any]:
    """
    Check for synonyms (e.g. force = True -> yes = True)
    """
    if args_dict.get('force', None):
        args_dict['yes'] = True
    if args_dict.get('async', None):
        args_dict['unblock'] = True
    if args_dict.get('mrsm_instance', None):
        args_dict['instance'] = args_dict['mrsm_instance']
    if args_dict.get('skip_check_existing', None):
        args_dict['check_existing'] = False
    return args_dict

def parse_dict_to_sysargs(
        args_dict : Dict[str, Any]
    ) -> List[str]:
    """
    Revert an arguments dictionary back to a command line list.
    """
    import json
    from meerschaum.actions.arguments._parser import get_arguments_triggers
    sysargs = []
    sysargs += args_dict.get('action', [])
    allow_none_args = {'location_keys'}

    triggers = get_arguments_triggers()

    for a, t in triggers.items():
        if a == 'action' or a not in args_dict:
            continue
        ### Add boolean flags
        if isinstance(args_dict[a], bool):
            if args_dict[a] is True:
                sysargs += [t[0]]
        else:
            ### Add list flags
            if (
                isinstance(args_dict[a], list) or isinstance(args_dict[a], tuple)
            ):
                if len(args_dict[a]) > 0:
                    sysargs += [t[0]] + list(args_dict[a])

            ### Add dict flags
            elif isinstance(args_dict[a], dict):
                if len(args_dict[a]) > 0:
                    sysargs += [t[0], json.dumps(args_dict[a])]

            ### Account for None and other values
            elif (args_dict[a] is not None) or (args_dict[a] is None and a in allow_none_args):
                sysargs += [t[0], args_dict[a]]

    return sysargs



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions to handle debug statements
"""

from __future__ import annotations
from meerschaum.utils.typing import Union, Optional, List

def dprint(
        msg : str,
        leader : bool = True,
        package: bool = True,
        color : Optional[Union[str, List[str]]] = None,
        attrs : Optional[List[str]] = None,
        nopretty : bool = False,
        **kw
    ) -> None:
    """
    Print a debug message.
    """
    if attrs is None:
        attrs = []
    if not isinstance(color, bool) and not nopretty:
        try:
            from meerschaum.utils.formatting import CHARSET, ANSI, colored
        except ImportError:
            CHARSET, ANSI, colored = 'ascii', False, None
        from meerschaum.config._paths import CONFIG_DIR_PATH, PERMANENT_PATCH_DIR_PATH
        from meerschaum.config import _config
        cf = _config('formatting')
        _color = color
    else:
        CHARSET, ANSI, colored, _color, cf = 'ascii', False, None, None, None

    import logging, sys, inspect
    logging.basicConfig(format='%(message)s')
    log = logging.getLogger(__name__)

    parent_frame = inspect.stack()[1][0]
    parent_info = inspect.getframeinfo(parent_frame)
    parent_lineno = parent_info.lineno
    parent_globals = parent_frame.f_globals
    parent_package = parent_globals['__name__']
    msg = str(msg)
    premsg = ""
    if package:
        premsg = parent_package + ':' + str(parent_lineno) + '\n'
    if leader and cf is not None:
        try:
            debug_leader = cf['formatting']['debug'][CHARSET]['icon'] if cf is not None else ''
        except KeyError:
            print(
                "Failed to load config. " +
                "Please delete the following directories and restart Meerschaum:"
            )
            for p in [CONFIG_DIR_PATH, PERMANENT_PATCH_DIR_PATH]:
                print('  - ' + str(p))
            debug_leader = ''
            ### crash if we can't load the leader
            #  sys.exit(1)
        premsg = ' ' + debug_leader + ' ' + premsg
    if ANSI:
        if _color is not None:
            if isinstance(_color, str):
                _color = [_color]
        else:
            if cf is not None and not nopretty:
                try:
                    _color = cf['formatting']['debug']['ansi']['rich'] if cf is not None else {}
                except KeyError:
                    _color = {}
            else:
                _color = []
        if colored is not None:
            premsg = colored(premsg, **_color)
    #  log.warning(premsg + msg, **kw)
    print(premsg + msg)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define a custom Thread class with a callback method
"""
import threading
class Thread(threading.Thread):
    """
    Wrapper for threading.Thread with optional callback and error_callback functions
    """

    def __init__(self, *args, callback=None, error_callback=None, **kw):
        target = kw.pop('target')
        super().__init__(target=self.wrap_target_with_callback, *args, **kw)
        self.callback = callback
        self.error_callback = error_callback
        self.method = target
        self._return = None

    def wrap_target_with_callback(self, *args, **kw):
        """
        Wrap the designated target function with a try-except.
        Captures the output and executes either the callback or error_callback.
        """
        try:
            result = self.method(*args, **kw)
            success = True
        except Exception as e:
            success = False
            result = None

        cb = self.callback if success else self.error_callback
        if cb is not None:
            cb(result)

    def join(self):
        """
        Return the thread's return value upon joining.
        """
        threading.Thread.join(self)
        return self._return

    def run(self):
        """
        Set the return to the result of the target.
        """
        self._return = self._target(*self._args, **self._kwargs)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Register a Pipe object
"""

def register(
        self,
        debug : bool = False
    ):
    """
    Send a POST to the Meerschaum API to register a new Pipe
    """
    if not self.parameters:
        self.parameters = {
            'columns' : self.columns,
        }
    return self.instance_connector.register_pipe(self, debug=debug)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Execute Meerschaum Actions via the API
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple

from meerschaum.api import fastapi, app, endpoints, get_api_connector, debug, manager
from meerschaum.actions import actions
import meerschaum._internal.User
actions_endpoint = endpoints['actions']

@app.get(actions_endpoint)
def get_actions() -> list:
    """
    Return a list of available actions
    """
    return list(actions)

@app.post(actions_endpoint + "/{action}")
def do_action(
        action : str,
        keywords : dict = fastapi.Body(...),
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> SuccessTuple:
    """
    Perform a Meerschaum action (if permissions allow it).
    """
    if curr_user.type != 'admin':
        from meerschaum.config import get_config
        allow_non_admin = get_config('system', 'api', 'permissions', 'actions', 'non_admin', patch=True)
        if not allow_non_admin:
            return False, (
                "The administrator for this server has not allowed users to perform actions.\n\n" +
                "Please contact the system administrator, or if you are running this server, " +
                "open the configuration file with `edit config system` and search for 'permissions'. " +
                " Under the keys system:api:permissions:actions, you can allow non-admin users to perform actions."
        )

    if action not in actions:
        return False, f"Invalid action '{action}'."
    if 'mrsm_instance' not in keywords:
        keywords['mrsm_instance'] = str(get_api_connector())
    _debug = keywords.get('debug', debug)
    if 'debug' in keywords:
        del keywords['debug']
    return actions[action](debug=_debug, **keywords)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Routes for managing plugins
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, List, SuccessTuple, Union

from meerschaum.api import (
    fastapi,
    app,
    endpoints,
    get_api_connector,
    pipes,
    get_pipe,
    manager,
    debug,
)
from meerschaum.api.tables import get_tables
from fastapi import FastAPI, File, UploadFile
from meerschaum.utils.packages import attempt_import
import meerschaum._internal.User
from meerschaum._internal.Plugin import Plugin
starlette_responses = attempt_import('starlette.responses', warn=False)
FileResponse = starlette_responses.FileResponse

sqlalchemy = attempt_import('sqlalchemy')
plugins_endpoint = endpoints['plugins']

@app.post(plugins_endpoint + '/{name}')
def register_plugin(
        name : str,
        version : str = None,
        attributes : str = None,
        archive : UploadFile = File(...),
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> SuccessTuple:
    """
    Register a plugin and save its archive file.
    """
    from meerschaum.config import get_config
    allow_plugins = get_config('system', 'api', 'permissions', 'registration', 'plugins')
    if not allow_plugins:
        return False, (
            "The administrator for this server has not allowed plugin registration.\n\n" +
            "Please contact the system administrator, or if you are running this server, " +
            "open the configuration file with `edit config system` and search for 'permissions'. " +
            " Under the keys `api:permissions:registration`, " +
            "you can toggle various registration types."
        )

    import json, shutil, pathlib, os
    get_tables()
    if attributes is None:
        attributes = json.dumps({})
    attributes = json.loads(attributes)

    plugin = Plugin(name, version=version, attributes=attributes)
    plugin_user_id = get_api_connector().get_plugin_user_id(plugin)
    curr_user_id = get_api_connector().get_user_id(curr_user)
    if plugin_user_id is not None and plugin_user_id != curr_user_id:
        return False, f"User '{curr_user.username}' cannot edit plugin '{plugin}'."
    plugin.user_id = curr_user_id

    success, msg = get_api_connector().register_plugin(plugin, make_archive=False, debug=debug)

    ### TODO delete and install new version of plugin on success
    if success:
        archive_path = plugin.archive_path
        temp_archive_path = pathlib.Path(str(archive_path) + '.tmp')
        with temp_archive_path.open('wb') as buff:
            shutil.copyfileobj(archive.file, buff)
        temp_archive_path.chmod(0o775)
        os.rename(temp_archive_path, archive_path)

    return success, msg

@app.get(plugins_endpoint + '/{name}')
def get_plugin(
        name : str
    ) -> Union[FileResponse, SuccessTuple]:
    """
    Download a plugin's archive file
    """
    plugin = Plugin(name)
    if plugin.archive_path.exists():
        return FileResponse(plugin.archive_path, filename=f'{plugin.name}.tar.gz')
    return False, f"Archive for plugin '{plugin}' could not be found."

@app.get(plugins_endpoint + '/{name}/attributes')
def get_plugin_attributes(
        name : str
    ) -> dict:
    """
    Get a plugin's attributes.
    """
    return get_api_connector().get_plugin_attributes(Plugin(name))

@app.get(plugins_endpoint)
def get_plugins(
        user_id : Optional[int] = None,
        search_term : Optional[str] = None,
    ) -> List[str]:
    """
    Return a list of registered plugins.
    """
    return get_api_connector().get_plugins(user_id=user_id, search_term=search_term)

@app.delete(plugins_endpoint + '/{name}')
def delete_plugin(
        name : str,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> SuccessTuple:
    """
    Delete a plugin and its archive file from the repository.
    """
    get_tables()
    plugin = Plugin(name)
    plugin_user_id = get_api_connector().get_plugin_user_id(plugin)
    if plugin_user_id is None:
        return False, f"Plugin '{plugin}' is not registered."

    curr_user_id = get_api_connector().get_user_id(curr_user)
    if plugin_user_id != curr_user_id:
        return False, f"User '{curr_user.username}' cannot delete plugin '{plugin}'."
    plugin.user_id = curr_user_id

    _remove_success = plugin.remove()
    if not _remove_success[0]:
        return _remove_success

    _delete_success = get_api_connector().delete_plugin(plugin, debug=debug)
    if not _delete_success[0]:
        return _delete_success

    return True, f"Successfully deleted plugin '{plugin}'."



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define syncing components.
"""

from __future__ import annotations
from meerschaum.utils.packages import attempt_import
dash = attempt_import('dash', lazy=False)
dbc = attempt_import('dash_bootstrap_components', lazy=False)
dcc = attempt_import('dash_core_components', warn=False)
html = attempt_import('dash_html_components', warn=False)





#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for fetching new data into the Pipe
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, Any

def fetch(
        self,
        begin : Optional[datetime.datetime] = None,
        end : Optional[datetime.datetime] = None,
        sync_chunks : bool = False,
        debug : bool = False,
        **kw : Any
    ) -> 'pd.DataFrame or None':
    """
    Fetch a Pipe's latest data from its connector.

    returns : pd.DataFrame of newest unseen data
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn
    if 'fetch' in dir(self.connector):
        if self.connector.type == 'plugin':
            from meerschaum.utils.packages import activate_venv, deactivate_venv
            activate_venv(self.connector.label, debug=debug)
        df = self.connector.fetch(
            self,
            begin = begin,
            end = end,
            chunk_hook = (self.sync if sync_chunks else None),
            debug = debug,
            **kw
        )
        if self.connector.type == 'plugin':
            deactivate_venv(self.connector.label, debug=debug)
        ### Return True if we're syncing in parallel, else continue as usual.
        if sync_chunks:
            return True
        return df
    warn(f"No `fetch()` function defined for connector '{self.connector}'")
    return None



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Interact with Pipes metadata via SQLConnector.
"""
from __future__ import annotations
from meerschaum.utils.typing import (
    Union, Any, Sequence, SuccessTuple, Mapping, Tuple, Dict, Optional, List
)

def register_pipe(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False,
    ) -> SuccessTuple:
    """
    Register a new pipe.
    A pipe's attributes must be set before registering.
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.packages import attempt_import

    ### ensure pipes table exists
    from meerschaum.connectors.sql.tables import get_tables
    pipes = get_tables(mrsm_instance=self, debug=debug)['pipes']

    if pipe.get_id(debug=debug) is not None:
        return False, f"Pipe '{pipe}' is already registered"

    ### NOTE: if `parameters` is supplied in the Pipe constructor,
    ###       then `pipe.parameters` will exist and not be fetched from the database.

    ### 1. Prioritize the Pipe object's `parameters` first.
    ###    E.g. if the user manually sets the `parameters` property
    ###    or if the Pipe already exists
    ###    (which shouldn't be able to be registered anyway but that's an issue for later).
    parameters = None
    try:
        parameters = pipe.parameters
    except Exception as e:
        if debug:
            dprint(str(e))
        parameters = None

    ### ensure `parameters` is a dictionary
    if parameters is None:
        parameters = dict()

    import json
    sqlalchemy = attempt_import('sqlalchemy')
    values = {
        'connector_keys' : pipe.connector_keys,
        'metric_key'     : pipe.metric_key,
        'location_key'   : pipe.location_key,
        'parameters'     : json.dumps(parameters),
    }
    query = sqlalchemy.insert(pipes).values(**values)
    result = self.exec(query, debug=debug)
    if result is None:
        return False, f"Failed to register pipe '{pipe}'"
    return True, f"Successfully registered pipe '{pipe}'"

def edit_pipe(
        self,
        pipe : meerschaum.Pipe = None,
        patch : bool = False,
        debug : bool = False,
        **kw : Any
    ) -> SuccessTuple:
    """
    Edit a Pipe's parameters.

    :param pipe:
        The pipe to be edited.

    :param patch:
        If patch is True, update the existing parameters by cascading.
        Otherwise overwrite the parameters (default).

    :param debug: Verbosity toggle.
    """

    if pipe.id is None:
        return False, f"pipe '{pipe}' is not registered and cannot be edited."

    from meerschaum.utils.debug import dprint
    from meerschaum.utils.packages import attempt_import
    if not patch:
        parameters = pipe.parameters
    else:
        from meerschaum import Pipe
        from meerschaum.config._patch import apply_patch_to_config
        original_parameters = Pipe(
            pipe.connector_keys, pipe.metric_key, pipe.location_key
        ).parameters
        parameters = apply_patch_to_config(
            original_parameters,
            pipe.parameters
        )

    ### ensure pipes table exists
    from meerschaum.connectors.sql.tables import get_tables
    pipes = get_tables(mrsm_instance=self, debug=debug)['pipes']

    import json
    sqlalchemy = attempt_import('sqlalchemy')

    values = {
        'parameters' : json.dumps(parameters),
    }
    q = sqlalchemy.update(pipes).values(**values).where(
        pipes.c.pipe_id == pipe.id
    )

    result = self.exec(q, debug=debug)
    message = (
        f"Successfully edited pipe '{pipe}'."
        if result is not None else f"Failed to edit pipe '{pipe}'"
    )
    if result is None:
        message = f"Failed to edit pipe '{pipe}'"
    return (result is not None), message

def fetch_pipes_keys(
        self,
        connector_keys : Optional[List[str]] = None,
        metric_keys : Optional[List[str]] = None,
        location_keys : Optional[List[str]] = None,
        params : Optional[Dict[str, Any]] = None,
        debug : bool = False
    ) -> Optional[List[Tuple[str, str, Optional[str]]]]:
    """
    Return a list of tuples corresponding to the parameters provided.

    :param connector_keys:
        List of connector_keys to search by.

    :param metric_keys:
        List of metric_keys to search by.

    :param location_keys:
        List of location_keys to search by.

    :param params:
        Dictionary of additional parameters to search by.
        E.g. --params pipe_id:1

    :param debug: Verbosity toggle.
    """
    from meerschaum.utils.warnings import error
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.packages import attempt_import
    sqlalchemy = attempt_import('sqlalchemy')
    import json

    if connector_keys is None:
        connector_keys = []
    if metric_keys is None:
        metric_keys = []
    if location_keys is None:
        location_keys = []
    else:
        location_keys = [
            (lk if lk not in ('[None]', 'None', 'null') else None)
                for lk in location_keys
        ]


    if params is None:
        params = {}

    ### Add three primary keys to params dictionary
    ###   (separated for convenience of arguments).
    cols = {
        'connector_keys' : connector_keys,
        'metric_key' : metric_keys,
        'location_key' : location_keys,
    }

    ### Make deep copy so we don't mutate this somewhere else.
    from copy import deepcopy
    parameters = deepcopy(params)
    for col, vals in cols.items():
        ### Allow for IS NULL to be declared as a single-item list ([None]).
        if vals == [None]:
            vals = None
        if vals not in [[], ['*']]:
            parameters[col] = vals
    cols = {k : v for k, v in cols.items() if v != [None]}

    from meerschaum.connectors.sql.tables import get_tables
    pipes = get_tables(mrsm_instance=self, debug=debug)['pipes']

    _params = {}
    for k, v in parameters.items():
        _v = json.dumps(v) if isinstance(v, dict) else v
        _params[k] = _v

    ### parse regular params
    _where = [
        pipes.c[key] == val 
        for key, val in _params.items()
            if not isinstance(val, (list, tuple))
    ]
    q = sqlalchemy.select(
        [pipes.c.connector_keys, pipes.c.metric_key, pipes.c.location_key]
    ).where(sqlalchemy.and_(*_where))

    ### Parse IN params and add OR IS NULL if None in list.
    for c, vals in cols.items():
        if isinstance(vals, (list, tuple)) and vals:
            q = (
                q.where(pipes.c[c].in_(vals)) if None not in vals
                else q.where(sqlalchemy.or_(pipes.c[c].in_(vals), pipes.c[c].is_(None)))
            )

    ### execute the query and return a list of tuples
    try:
        if debug:
            dprint(q)
        rows = self.engine.execute(q).fetchall()
        return [(row['connector_keys'], row['metric_key'], row['location_key']) for row in rows]
    except Exception as e:
        error(str(e))

    return None

def create_indices(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False
    ) -> bool:
    """
    Create indices for a Pipe's datetime and ID columns.
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.connectors.sql._tools import sql_item_name, get_distinct_col_count
    from meerschaum.utils.warnings import warn
    index_queries = dict()

    if debug:
        dprint(f"Creating indices for Pipe '{pipe}'...")
    if (
        pipe.columns is None or
        pipe.columns.get('datetime', None) is None
    ):
        warn(f"Unable to create indices for pipe '{pipe}' without columns.")
        return False

    _datetime = pipe.get_columns('datetime', error=False)
    _datetime_name = sql_item_name(_datetime, self.flavor) if _datetime is not None else None
    _id = pipe.get_columns('id', error=False)
    _id_name = sql_item_name(_id, self.flavor) if _id is not None else None
    _pipe_name = sql_item_name(str(pipe), self.flavor)

    ### create datetime index
    if _datetime is not None:
        if self.flavor == 'timescaledb':
            _id_count = (
                get_distinct_col_count(_id, f"SELECT {_id_name} FROM {_pipe_name}", self)
                if _id is not None else None
            )
            dt_query = (
                f"SELECT create_hypertable('{_pipe_name}', " +
                f"'{_datetime}', " + (f"'{_id}', {_id_count}, " if _id is not None else '') +
                "migrate_data => true);"
            )
        elif self.flavor == 'postgresql':
            dt_query = f"CREATE INDEX ON {_pipe_name} ({_datetime_name})"
        elif self.flavor in ('mysql', 'mariadb'):
            dt_query = f"CREATE INDEX ON {_pipe_name} ({_datetime_name})"
        else: ### mssql, sqlite, etc.
            dt_query = f"CREATE INDEX {sql_item_name(_datetime + '_index', self.flavor)} ON {_pipe_name} ({_datetime_name})"

        index_queries[_datetime] = dt_query

    ### create id index
    if _id_name is not None:
        if self.flavor == 'timescaledb':
            ### Already created indices via create_hypertable.
            id_query = None
            pass
        elif self.flavor in ('postgresql'):
            id_query = f"CREATE INDEX ON {_pipe_name} ({_id_name})"
        elif self.flavor in ('mysql', 'mariadb'):
            id_query = f"CREATE INDEX ON {_pipe_name} ({_id_name})"
        else: ### mssql, sqlite, etc.
            id_query = f"CREATE INDEX {sql_item_name(_id + '_index', self.flavor)} ON {_pipe_name} ({_id_name})"

        if id_query is not None:
            index_queries[_id] = id_query

    failures = 0
    for col, q in index_queries.items():
        if debug:
            dprint(f"Creating index on column '{col}' for pipe '{pipe}'...")
        index_result = self.exec(q, silent=debug, debug=debug)
        if not index_result:
            failures += 1
    return failures == 0

def delete_pipe(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False,
    ) -> SuccessTuple:
    """
    Delete a Pipe's registration and drop its table.
    """
    from meerschaum.utils.warnings import warn
    from meerschaum.connectors.sql._tools import sql_item_name
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.packages import attempt_import
    sqlalchemy = attempt_import('sqlalchemy')

    ### try dropping first
    drop_tuple = pipe.drop(debug=debug)
    if not drop_tuple[0]:
        return drop_tuple

    if not pipe.id:
        return False, f"Pipe '{pipe}' is not registered"

    ### ensure pipes table exists
    from meerschaum.connectors.sql.tables import get_tables
    pipes = get_tables(mrsm_instance=self, debug=debug)['pipes']

    q = sqlalchemy.delete(pipes).where(pipes.c.pipe_id == pipe.id)
    if not self.exec(q, debug=debug):
        return False, f"Failed to delete registration for '{pipe}'"

    return True, "Success"

def get_backtrack_data(
        self,
        pipe : Optional[meerschaum.Pipe] = None,
        backtrack_minutes : int = 0,
        begin : Optional[datetime.datetime] = None,
        chunksize : Optional[int] = -1,
        debug : bool = False
    ) -> Optional[pandas.DataFrame]:
    """
    Get the most recent backtrack_minutes' worth of data from a Pipe.
    """
    from meerschaum.utils.warnings import error, warn
    if pipe is None:
        error("Pipe must be provided.")
    from meerschaum.utils.debug import dprint
    from meerschaum.connectors.sql._tools import dateadd_str
    if begin is None:
        begin = pipe.sync_time
    da = dateadd_str(
        flavor = self.flavor,
        datepart = 'minute',
        number = (-1 * backtrack_minutes),
        begin = begin
    )

    ### check for capitals
    from meerschaum.connectors.sql._tools import sql_item_name
    table = sql_item_name(str(pipe), self.flavor)
    dt = sql_item_name(pipe.get_columns('datetime'), self.flavor)

    query = f"SELECT * FROM {table}" + (f" WHERE {dt} > {da}" if da else "")

    df = self.read(query, chunksize=chunksize, debug=debug)

    if self.flavor == 'sqlite':
        from meerschaum.utils.misc import parse_df_datetimes
        df = parse_df_datetimes(df, debug=debug)

    return df

def get_pipe_data(
        self,
        pipe : Optional[meerschaum.Pipe] = None,
        begin : Union[datetime.datetime, str, None] = None,
        end : Union[datetime.datetime, str, None] = None,
        params : Optional[Dict[str, Any]] = None,
        debug : bool = False,
        **kw : Any
    ) -> Optional[pandas.DataFrame]:
    """
    Access a pipe's data from the SQL instance.

    :param begin:
        Lower bound for the query (inclusive)

    :param end:
        Upper bound for the query (inclusive)
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.connectors.sql._tools import sql_item_name, dateadd_str
    from meerschaum.utils.packages import import_pandas
    pd = import_pandas()

    query = f"SELECT * FROM {sql_item_name(str(pipe), self.flavor)}"
    where = ""

    dt = sql_item_name(pipe.get_columns('datetime'), self.flavor)

    if begin is not None:
        begin_da = dateadd_str(
            flavor = self.flavor,
            datepart = 'minute',
            number = 0,
            begin = begin
        )
        where += f"{dt} >= {begin_da}" + (" AND " if end is not None else "")

    if end is not None:
        end_da = dateadd_str(
            flavor = self.flavor,
            datepart = 'minute',
            number = 0,
            begin = end
        )
        where += f"{dt} <= {end_da}"

    if params is not None:
        from meerschaum.connectors.sql._tools import build_where
        where += build_where(params, self).replace('WHERE', ('AND' if (begin is not None or end is not None) else ""))

    if len(where) > 0:
        query += "\nWHERE " + where

    query += "\nORDER BY " + dt + " DESC"

    if debug:
        dprint(f"Getting Pipe data with begin = '{begin}' and end = '{end}'")
    df = self.read(query, debug=debug, **kw)
    if self.flavor == 'sqlite':
        from meerschaum.utils.misc import parse_df_datetimes
        ### NOTE: we have to consume the iterator here to ensure that datatimes are parsed correctly.
        df = parse_df_datetimes(df, debug=debug) if isinstance(df, pd.DataFrame) else (
            [parse_df_datetimes(c, debug=debug) for c in df]
        )
    return df

def get_pipe_id(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False,
    ) -> int:
    """
    Get a Pipe's ID from the pipes table.
    """
    from meerschaum.utils.packages import attempt_import
    import json
    sqlalchemy = attempt_import('sqlalchemy')
    from meerschaum.connectors.sql.tables import get_tables
    pipes = get_tables(mrsm_instance=self, debug=debug)['pipes']

    query = sqlalchemy.select([pipes.c.pipe_id]).where(
        pipes.c.connector_keys == pipe.connector_keys
    ).where(
        pipes.c.metric_key == pipe.metric_key
    ).where(
        pipes.c.location_key == pipe.location_key
    )
    _id = self.value(query, debug=debug)
    if _id is not None:
        _id = int(_id)
    return _id

def get_pipe_attributes(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False
    ) -> Optional[Mapping[Any, Any]]:
    """
    Get a Pipe's attributes dictionary
    """
    from meerschaum.utils.warnings import warn
    from meerschaum.connectors.sql.tables import get_tables
    from meerschaum.utils.packages import attempt_import
    sqlalchemy = attempt_import('sqlalchemy')
    pipes = get_tables(mrsm_instance=self, debug=debug)['pipes']

    if pipe.id is None:
        return None

    try:
        q = sqlalchemy.select([pipes]).where(pipes.c.pipe_id == pipe.id)
        attributes = dict(self.exec(q, debug=debug).first())
    except Exception as e:
        warn(e)
        return None

    ### handle non-PostgreSQL databases (text vs JSON)
    if not isinstance(attributes.get('parameters', None), dict):
        try:
            import json
            attributes['parameters'] = json.loads(attributes['parameters'])
        except Exception as e:
            attributes['parameters'] = dict()

    return attributes

def sync_pipe(
        self,
        pipe : meerschaum.Pipe.Pipe,
        df : Union[pandas.DataFrame, str, Dict[Any, Any]] = None,
        begin : Optional[datetime.datetime] = None,
        end : Optional[datetime.datetime] = None,
        chunksize : Optional[int] = -1,
        check_existing : bool = True,
        blocking : bool = True,
        debug : bool = False,
        **kw : Any
    ) -> SuccessTuple:
    """
    Sync a pipe using a SQL Connection.

    :param pipe:
        The Meerschaum Pipe instance into which to sync the data.

    :param df:
        An optional DataFrame to sync into the pipe.
        Defaults to `None`.

    :param begin:
        Optionally specify the earliest datetime to search for data.
        Defaults to `None`.

    :param end:
        Optionally specify the latest datetime to search for data.
        Defaults to `None`.

    :param chunksize:
        Specify the number of rows to sync per chunk.
        If -1, resort to system configuration (default is 900).
        A `chunksize` of `None` will sync all rows in one transaction.
        Defaults to -1.

    :param check_existing:
        If True, pull and diff with existing data from the pipe, defaults to True.

    :param blocking:
        If True, wait for sync to finish and return its result, otherwise asyncronously sync.
        Defaults to True.

    :param debug:
        Verbosity toggle. Defaults to False.

    :param kw:
        Catch-all for keyword arguments.
    """
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.packages import import_pandas
    if df is None:
        msg = f"DataFrame is None. Cannot sync pipe '{pipe}'."
        warn(msg)
        return False, msg

    ## allow syncing for JSON or dict as well as DataFrame
    pd = import_pandas()
    if isinstance(df, dict):
        df = pd.DataFrame(df)
    elif isinstance(df, str):
        df = pd.read_json(df)

    ### if Pipe is not registered
    if not pipe.id:
        register_tuple = pipe.register(debug=debug)
        if not register_tuple[0]:
            return register_tuple

    ### quit here if implicitly syncing MQTT pipes.
    ### (pipe.sync() is called in the callback of the MQTTConnector.fetch() method)
    if df is None and pipe.connector.type == 'mqtt':
        return True, "Success"

    ### df is the dataframe returned from the remote source
    ### via the connector
    if debug:
        dprint("Fetched data:\n" + str(df))

    ### if table does not exist, create it with indices
    is_new = False
    if not pipe.exists(debug=debug):
        check_existing = False
        is_new = True

    new_data_df = filter_existing(pipe, df, chunksize=chunksize, debug=debug) if check_existing else df
    if debug:
        dprint("New unseen data:\n" + str(new_data_df))

    if_exists = kw.get('if_exists', 'append')
    if 'if_exists' in kw:
        kw.pop('if_exists')
    if 'name' in kw:
        kw.pop('name')

    ### append new data to Pipe's table
    result = self.to_sql(
        new_data_df,
        name = str(pipe),
        if_exists = if_exists,
        debug = debug,
        as_tuple = True,
        chunksize = chunksize,
        **kw
    )
    if is_new:
        if not self.create_indices(pipe, debug=debug):
            if debug:
                dprint(f"Failed to create indices for pipe '{pipe}'. Continuing...")
    return result


def filter_existing(pipe, df, chunksize : Optional[int] = -1, debug : bool = False):
    """
    Remove duplicate data from backtrack_data and a new dataframe
    """
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.misc import round_time
    from meerschaum.utils.packages import attempt_import
    np = attempt_import('numpy')
    import datetime as datetime_pkg
    ### begin is the oldest data in the new dataframe
    try:
        min_dt = df[pipe.get_columns('datetime')].min().to_pydatetime()
    except:
        min_dt = pipe.get_sync_time(debug=debug)
    if min_dt in (np.nan, None):
        min_dt = None
    begin = round_time(
        min_dt,
        to = 'down'
    ) - datetime_pkg.timedelta(minutes=1)
    if debug:
        dprint(f"Looking at data newer than '{begin}'.")

    ### backtrack_df is existing Pipe data that overlaps with the fetched df
    try:
        backtrack_minutes = pipe.parameters['fetch']['backtrack_minutes']
    except Exception as e:
        backtrack_minutes = 0

    backtrack_df = pipe.get_backtrack_data(
        begin = begin,
        backtrack_minutes = backtrack_minutes,
        chunksize = chunksize,
        debug = debug
    )
    if debug:
        dprint("Existing data:\n" + str(backtrack_df))

    ### remove data we've already seen before
    from meerschaum.utils.misc import filter_unseen_df
    return filter_unseen_df(backtrack_df, df, debug=debug)

def get_sync_time(
        self,
        pipe : 'meerschaum.Pipe',
        params : dict = None,
        debug : bool = False,
    ) -> 'datetime.datetime':
    """
    Get a Pipe's most recent datetime
    """
    from meerschaum.connectors.sql._tools import sql_item_name, build_where
    from meerschaum.utils.warnings import warn
    import datetime
    table = sql_item_name(str(pipe), self.flavor)
    dt = sql_item_name(pipe.get_columns('datetime'), self.flavor)

    where = "" if params is None else build_where(params, self)
    q = f"SELECT {dt}\nFROM {table}{where}\nORDER BY {dt} DESC\nLIMIT 1"
    if self.flavor == 'mssql':
        q = f"SELECT TOP 1 {dt}\nFROM {table}{where}\nORDER BY {dt} DESC"
    try:
        from meerschaum.utils.misc import round_time
        import datetime
        db_time = self.value(q, silent=True, debug=debug)

        ### No datetime could be found.
        if db_time is None:
            return None
        ### sqlite returns str.
        elif isinstance(db_time, str):
            from meerschaum.utils.packages import attempt_import
            dateutil_parser = attempt_import('dateutil.parser')
            st = dateutil_parser.parse(db_time)
        ### Do nothing if a datetime object is returned.
        elif isinstance(db_time, datetime.datetime):
            st = db_time
        ### Convert pandas timestamp to Python datetime.
        else:
            st = db_time.to_pydatetime()

        ### round down to smooth timestamp
        sync_time = round_time(st, date_delta=datetime.timedelta(minutes=1), to='down')

    except Exception as e:
        sync_time = None
        warn(str(e))

    return sync_time

def pipe_exists(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False
    ) -> bool:
    """
    Check that a Pipe's table exists.
    """
    from meerschaum.connectors.sql._tools import table_exists
    exists = table_exists(str(pipe), self, debug=debug)
    if debug:
        from meerschaum.utils.debug import dprint
        dprint(f"Pipe '{pipe}' " + ('exists.' if exists else 'does not exist.'))
    return exists

def get_pipe_rowcount(
        self,
        pipe : 'meerschaum.Pipe',
        begin : 'datetime.datetime' = None,
        end : 'datetime.datetime' = None,
        remote : bool = False,
        params : Optional[Dict[str, Any]] = None,
        debug : bool = False
    ) -> int:
    """
    Return the number of rows between datetimes for a Pipe's instance cache or remote source
    """
    from meerschaum.connectors.sql._tools import dateadd_str, sql_item_name
    from meerschaum.utils.warnings import error, warn
    if remote:
        msg = f"'fetch:definition' must be an attribute of pipe '{pipe}' to get a remote rowcount"
        if 'fetch' not in pipe.parameters:
            error(msg)
            return None
        if 'definition' not in pipe.parameters['fetch']:
            error(msg)
            return None

    _pipe_name = sql_item_name(str(pipe), self.flavor)
    _datetime_name = sql_item_name(pipe.get_columns('datetime'), self.flavor)

    src = (
        f"SELECT {_datetime_name} FROM {_pipe_name}"
        if not remote else pipe.parameters['fetch']['definition']
    )
    query = f"""
    WITH src AS ({src})
    SELECT COUNT({_datetime_name})
    FROM src
    """
    if begin is not None or end is not None:
        query += "WHERE"
    if begin is not None:
        query += f"""
        {_datetime_name} > {dateadd_str(flavor=self.flavor, datepart='minute', number=(0), begin=begin)}
        """
    if end is not None and begin is not None:
        query += "AND"
    if end is not None:
        query += f"""
        {_datetime_name} <= {dateadd_str(flavor=self.flavor, datepart='minute', number=(0), begin=end)}
        """
    if params is not None:
        from meerschaum.connectors.sql._tools import build_where
        query += build_where(params, self).replace('WHERE', (
            'AND' if (begin is not None or end is not None)
                else 'WHERE'
            )
        )
        
    result = self.value(query, debug=debug)
    try:
        return int(result)
    except Exception as e:
        return None

def drop_pipe(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False
    ) -> SuccessTuple:
    """
    Drop a pipe's tables but maintain its registration.
    """
    if not pipe.exists(debug=debug):
        return True, f"Pipe '{pipe}' does not exist, so nothing was dropped."

    from meerschaum.connectors.sql._tools import sql_item_name
    pipe_name = sql_item_name(str(pipe), self.flavor)
    success = self.exec(f"DROP TABLE {pipe_name}", silent=True, debug=debug) is not None
    if not success:
        success = self.exec(f"DROP VIEW {pipe_name}", silent=True, debug=debug) is not None
    
    msg = "Success" if success else f"Failed to drop pipe '{pipe}'."
    return success, msg

def get_pipe_table(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False,
    ) -> sqlalchemy.Table:
    """
    Return a sqlalchemy Table for a Pipe bound to this SQLConnector's engine.
    """
    from meerschaum.connectors.sql._tools import get_sqlalchemy_table
    return get_sqlalchemy_table(str(pipe), connector=self, debug=debug)

def get_pipe_columns_types(
        self,
        pipe : meerschaum.Pipe,
        debug : bool = False,
    ) -> Optional[Dict[str, str]]:
    """
    Return a dictionary of a pipe's table's columns to their data types.

    E.g. An example dictionary for a small table.

    ```
    >>> {
    ...   'dt': 'TIMESTAMP WITHOUT TIMEZONE',
    ...   'id': 'BIGINT',
    ...   'val': 'DOUBLE PRECISION',
    ... }
    >>> 
    ```
    """
    table_columns = {}
    try:
        pipe_table = self.get_pipe_table(pipe, debug=debug)
        for col in pipe_table.columns:
            table_columns[str(col.name)] = str(col.type)
    except Exception as e:
        from meerschaum.utils.warnings import warn
        warn(e)
        table_columns = None

    return table_columns



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Miscellaneous routes
"""

import os
from meerschaum.api import app, endpoints, check_allow_chaining
from starlette.responses import FileResponse

@app.get(endpoints['favicon'])
def get_favicon() -> FileResponse:
    from meerschaum.config._paths import API_STATIC_PATH
    return FileResponse(os.path.join(API_STATIC_PATH, 'ico', 'logo.ico'))

@app.get(endpoints['chaining'])
def get_chaining_status() -> bool:
    """
    Return whether this API instance allows chaining.
    """
    return check_allow_chaining()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for running the Docker Compose stack
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Any, List, Optional

def stack(
        action : Optional[List[str]] = None,
        sysargs : Optional[List[str]] = None,
        sub_args : Optional[List[str]] = None,
        yes : bool = False,
        noask : bool = False,
        force : bool = False,
        debug : bool = False,
        **kw : Any
    ) -> SuccessTuple:
    """
    Control the Meerschaum stack with Docker Compose.
    Usage: `stack {command}`
    
    command : action[0] : default 'up'
        Docker Compose command to run. E.g. 'config' will print Docker Compose configuration
    """
    from subprocess import call
    import meerschaum.config.stack
    from meerschaum.config.stack import get_necessary_files, write_stack
    from meerschaum.config._paths import STACK_COMPOSE_PATH
    #  from meerschaum.utils.packages import reload_package
    from meerschaum.utils.prompt import yes_no
    import meerschaum.config
    from meerschaum.utils.packages import attempt_import, run_python_package, venv_contains_package
    from meerschaum.config import get_config
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.formatting import ANSI
    import os, sys

    if action is None:
        action = []
    if sysargs is None:
        sysargs = []
    if sub_args is None:
        sub_args = []

    bootstrap_question = (
        "Bootstrap configuration?\n\n"
        "The following files will be overwritten:"
    )
    for f in get_necessary_files():
        bootstrap_question += "\n  - " + str(f)
    bootstrap_question += "\n"

    bootstrap = False
    for fp in get_necessary_files():
        if not os.path.isfile(fp):
            if not force:
                if yes_no(bootstrap_question, yes=yes, noask=noask):
                    bootstrap = True
                else:
                    warn_message = "Cannot start stack without bootstrapping"
                    warn(warn_message)
                    return False, warn_message
            else: ### force is True
                bootstrap = True
            break
    ### if bootstrap flag was set, create files
    if bootstrap:
        write_stack(debug=debug)

    compose_command = ['up']
    ### default: alias stack as docker-compose
    if len(action) > 0 and action[0] != '':
        compose_command = action

    ### define project name when starting containers
    project_name_list = ['--project-name', get_config('stack', 'project_name', patch=True, substitute=True)]
    
    ### disable ANSI if the user sets ANSI mode to False
    ansi_list = [] if ANSI else ['--no-ansi']

    debug_list = ['--log-level', 'DEBUG'] if debug else []

    ### prepend settings before the docker-compose action
    settings_list = project_name_list + ansi_list + debug_list

    _compose_venv = 'mrsm'
    compose = attempt_import('compose', lazy=False, venv=_compose_venv)
    docker = attempt_import('docker', warn=False)
    try:
        client = docker.from_env()
    except Exception as e:
        warn("Could not connect to Docker. Is the Docker service running?", stack=False)
        print(
            "To start the Docker service, run `sudo systemctl start docker` or `sudo dockerd`.\n"
            + "On Windows or MacOS, make sure Docker Desktop is running.",
            file = sys.stderr,
        )
        return False, "Failed to connect to the Docker engine."
    ### If docker-compose is installed globally, don't use the `mrsm` venv.
    if not venv_contains_package('compose', _compose_venv):
        _compose_venv = None
    cmd_list = settings_list + compose_command + (
        sysargs[2:] if (len(sysargs) > 2 and not sub_args) else sub_args
    )
    if debug:
        dprint(cmd_list)
    rc = run_python_package(
        'compose', args=cmd_list, cwd=STACK_COMPOSE_PATH.parent, venv=_compose_venv
    )

    return rc == 0, ("Success" if rc == 0 else f"Failed to execute commands: {cmd_list}")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Plugin metadata class
"""

from __future__ import annotations
from meerschaum.utils.typing import (
    Dict,
    List,
    Optional,
    Any,
    SuccessTuple,
    Union,
)
from meerschaum.config._paths import (
    PLUGINS_RESOURCES_PATH,
    PLUGINS_ARCHIVES_RESOURCES_PATH,
    PLUGINS_TEMP_RESOURCES_PATH
)
import os, pathlib
_tmpversion = None

class Plugin:
    def __init__(
        self,
        name : str,
        version : Optional[str] = None,
        user_id : Optional[int] = None,
        attributes : Optional[Dict[str, Any]] = None,
        archive_path : Optional[pathlib.Path] = None
    ):
        if attributes is None:
            attributes = {}
        self.name = name
        self.attributes = attributes
        self.user_id = user_id
        if version is None:
            try:
                self.version = self.module.__version__
            except Exception as e:
                self.version = None
        else:
            self.version = version

        if archive_path is None:
            self.archive_path = pathlib.Path(
                os.path.join(PLUGINS_ARCHIVES_RESOURCES_PATH, f'{self.name}.tar.gz')
            )
        else:
            self.archive_path = archive_path

    @property
    def module(self):
        """
        Return a Plugin's Python module.
        """
        if '_module' not in self.__dict__:
            from meerschaum.plugins import import_plugins
            self._module = import_plugins(str(self))
            #  from meerschaum.plugins import get_plugins_modules
            #  for m in get_plugins_modules():
                #  if self.name == m.__name__.split('.')[-1]:
                    #  self._module = m
                    #  break
        #  if '_module' not in self.__dict__:
            #  return None
        return self._module

    @property
    def __file__(self) -> Union[str, None]:
        if self.module is None:
            return None
        return self.module.__file__

    def make_tar(self, debug : bool = False) -> pathlib.Path:
        """
        Compress the plugin's source files into a `.tar.gz` archive and return the archive's path.
        """
        import tarfile
        from meerschaum.utils.debug import dprint

        old_cwd = os.getcwd()
        os.chdir(PLUGINS_RESOURCES_PATH)

        if self.__file__ is None:
            from meerschaum.utils.warnings import error
            error(f"Could not find file for plugin '{self}'.")
        if '__init__.py' in self.__file__:
            path = self.__file__.replace('__init__.py', '')
            is_dir = True
        else:
            path = self.__file__
            is_dir = False

        tarf = tarfile.open(
            self.archive_path,
            'w:gz'
        )

        patterns_to_ignore = {
            '.pyc',
            '__pycache__/',
            'eggs/',
            '__pypackages__/',
            '.git',
        }

        if not is_dir:
            tarf.add(f"{self.name}.py")
        else:
            for root, dirs, files in os.walk(self.name):
                for f in files:
                    good_file = True
                    fp = os.path.join(root, f)
                    for pattern in patterns_to_ignore:
                        if pattern in str(fp) or f.startswith('.'):
                            good_file = False
                            break
                    if good_file:
                        if debug:
                            dprint(f"Adding '{fp}'...")
                        tarf.add(fp)

        ### clean up and change back to old directory
        tarf.close()
        os.chdir(old_cwd)

        ### change to 775 to avoid permissions issues with the API in a Docker container
        self.archive_path.chmod(0o775)

        if debug:
            dprint(f"Created archive '{self.archive_path}'.")
        return self.archive_path

    def remove(
            self,        
            force : bool = False,
            debug : bool = False
        ) -> SuccessTuple:
        """
        Remove a plugin's archive file.
        """
        if not self.archive_path.exists():
            return True, f"Archive file for plugin '{self}' does not exist."
        try:
            self.archive_path.unlink()
        except Exception as e:
            return False, f"Failed to remove plugin '{self}'\n{e}"
        return True, "Success"

    def install(
            self,
            force : bool = False,
            debug : bool = False
        ) -> SuccessTuple:
        """
        Extract a plugin's tar archive to the plugins directory.
        This function checks if the plugin is already installed and if the version is equal or
        greater than the existing installation.
        """
        from meerschaum.utils.warnings import warn, error
        if debug:
            from meerschaum.utils.debug import dprint
        import tarfile, pathlib, shutil
        from meerschaum.plugins import reload_plugins
        old_cwd = os.getcwd()
        old_version = ''
        new_version = ''
        if not self.archive_path.exists():
            return False, f"Missing archive file for plugin '{self}'."
        is_installed = None
        if self.__file__ is not None:
            is_installed = True
            try:
                old_version = self.module.__version__
            except Exception as e:
                old_version = ''
            if debug:
                dprint(f"Found existing version '{old_version}' for plugin '{self}'.")
        try:
            tarf = tarfile.open(
                self.archive_path,
                'r:gz'
            )
        except Exception as e:
            warn(e)
            return False, f"Plugin '{self.name}' could not be downloaded."

        temp_dir = pathlib.Path(os.path.join(PLUGINS_TEMP_RESOURCES_PATH, self.name))
        temp_dir.mkdir(exist_ok=True)

        if debug:
            dprint(f"Extracting '{self.archive_path}' to '{temp_dir}'...")
        try:
            tarf.extractall(temp_dir)
        except Exception as e:
            success, msg = False, f"Failed to extract plugin '{self}'."

        ### search for version information
        files = os.listdir(temp_dir)
        
        if str(files[0]) == self.name:
            is_dir = True
        elif str(files[0]) == self.name + '.py':
            is_dir = False
        else:
            error(f"Unknown format encountered for plugin '{self}'.")

        fpath = pathlib.Path(os.path.join(temp_dir, files[0]))
        if is_dir:
            fpath = pathlib.Path(os.path.join(fpath, '__init__.py'))
        with open(fpath, 'r') as f:
            lines = f.readlines()
        global _tmpversion
        for l in lines:
            if '__version__' in l:
                _l = l.replace('__version__', '_tmpversion')
                exec(_l, globals())
                new_version = _tmpversion
                if debug:
                    dprint(f"Attempting to install plugin '{self}' version '{new_version}'...")
                break

        from meerschaum.utils.packages import attempt_import
        packaging_version = attempt_import('packaging.version')
        is_new_version = (
            packaging_version.parse(old_version) < packaging_version.parse(new_version)
        )
        is_same_version = (
            packaging_version.parse(old_version) == packaging_version.parse(new_version)
        )
        success_msg = f"Successfully installed plugin '{self}'."
        success, abort = None, None
        if is_same_version and not force:
            success, msg = True, (
                f"Plugin '{self}' is up-to-date (version {old_version}).\n\n" +
                "    Install again with `-f` or `--force` to reinstall."
            )
            abort = True
        elif is_new_version or force:
            for src_dir, dirs, files in os.walk(temp_dir):
                if success is not None:
                    break
                dst_dir = str(src_dir).replace(str(temp_dir), str(PLUGINS_RESOURCES_PATH))
                if not os.path.exists(dst_dir):
                    os.mkdir(dst_dir)
                for f in files:
                    src_file = os.path.join(src_dir, f)
                    dst_file = os.path.join(dst_dir, f)
                    if os.path.exists(dst_file):
                        os.remove(dst_file)

                    if debug:
                        dprint(f"Moving '{src_file}' to '{dst_dir}'...")
                    try:
                        shutil.move(src_file, dst_dir)
                    except Exception as e:
                        success, msg = False, (
                            f"Failed to install plugin '{self}': " +
                            f"Could not move file '{src_file}' to '{dst_dir}'"
                        )
                        print(msg)
                        break
            if success is None:
                success, msg = True, success_msg
        else:
            success, msg = False, (
                f"Your Existing version '{old_version}' is higher than attempted version " +
                f"'{new_version}'."
            )

        shutil.rmtree(temp_dir)
        tarf.close()
        os.chdir(old_cwd)

        ### Reload the plugin's module.
        if '_module' in self.__dict__:
            del self.__dict__['_module']
        reload_plugins([self.name], debug=debug)

        ### if we've already failed, return here
        if not success or abort:
            return success, msg

        ### attempt to install dependencies
        if not self.install_dependencies(debug=debug):
            return False, f"Failed to install dependencies for plugin '{self}'."

        ### handling success tuple, bool, or other (typically None)
        setup_tuple = self.setup(debug=debug)
        if isinstance(setup_tuple, tuple):
            if not setup_tuple[0]:
                success, msg = setup_tuple
        elif isinstance(setup_tuple, bool):
            if not setup_tuple:
                success, msg = False, (
                    f"Failed to run post-install setup for plugin '{self}'." + '\n' +
                    f"Check `setup()` in '{self.__file__}' for more information " +
                    f"(no error message provided)."
                )
            else:
                success, msg = True, success_msg
        elif setup_tuple is None:
            success = True
            msg = (
                f"Post-install for plugin '{self}' returned None. " +
                f"Assuming plugin successfully installed."
            )
            warn(msg)
        else:
            success = False
            msg = (
                f"Post-install for plugin '{self}' returned unexpected value " +
                f"of type '{type(setup_tuple)}': {setup_tuple}"
            )

        return success, msg

    def setup(self, *args : str, debug : bool = False, **kw : Any) -> SuccessTuple:
        """
        If exists, run the plugin's setup() function.
        """
        from meerschaum.utils.packages import activate_venv, deactivate_venv
        from meerschaum.utils.debug import dprint
        import inspect
        _setup = None
        for name, fp in inspect.getmembers(self.module):
            if name == 'setup' and inspect.isfunction(fp):
                _setup = fp
                break

        ### assume success if no setup() is found (not necessary)
        if _setup is None:
            return True

        sig = inspect.signature(_setup)
        has_debug, has_kw = ('debug' in sig.parameters), False
        for k, v in sig.parameters.items():
            if '**' in str(v):
                has_kw = True
                break

        _kw = {}
        if has_kw:
            _kw.update(kw)
        if has_debug:
            _kw['debug'] = debug

        if debug:
            dprint(f"Running setup for plugin '{self}'...")
        try:
            activate_venv(venv=self.name, debug=debug)
            return_tuple = _setup(*args, **_kw)
            deactivate_venv(venv=self.name, debug=debug)
        except Exception as e:
            return False, str(e)

        if isinstance(return_tuple, tuple):
            return return_tuple
        if isinstance(return_tuple, bool):
            return return_tuple, f"Setup for Plugin '{self.name}' did not return a message."
        if return_tuple is None:
            return False, f"Setup for Plugin '{self.name}' returned None."
        return False, f"Unknown return value from setup for Plugin '{self.name}': {return_tuple}"

    @property
    def dependencies(self) -> List[str]:
        """
        If the Plugin has specified dependencies in a list called `required`, return the list.
        """
        from meerschaum.utils.packages import activate_venv, deactivate_venv
        import inspect
        activate_venv(venv=self.name)
        required = []
        for name, val in inspect.getmembers(self.module):
            if name == 'required':
                required = val
                break
        deactivate_venv(venv=self.name)
        return required

    def install_dependencies(self, debug : bool = False) -> bool:
        """
        If specified, install dependencies.
        """
        from meerschaum.utils.packages import pip_install
        from meerschaum.utils.debug import dprint
        _deps = self.dependencies
        if _deps:
            if debug:
                dprint(f"Installing dependencies: {_deps}")
            return pip_install(*_deps, venv=self.name, debug=debug)
        return True

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)

    def __del__(self):
        pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Stack configuration for Grafana
"""

import os
from meerschaum.config.stack import db_host, db_user, db_pass, db_port, db_base

default_datasource = {
    'apiVersion' : 1,
    'datasources' : [
        {
            'name' : 'Meerschaum Main',
            'type' : 'postgres',
            'jsonData' : {
                'sslmode' : 'disable',
                'postgresVersion' : 1300,
                'timescaledb' : True,
            },
            'user' : db_user,
            'secureJsonData' : {
                'password' : db_pass,
            },
            'database' : db_base,
            'url' : db_host + ':' + str(db_port),
            'isDefault' : True,
            'editable' : True,
        }
    ],
}

default_dashboard = {
    'apiVersion' : 1,
    'providers' : [
        {
            'name' : 'Default',
            'folder' : 'Meerschaum Dashboards',
            'options' : {
                'path' : '/etc/grafana/provisioning/dashboards',
            },
        }
    ],
}

### build config dictionary
default_grafana_config = dict()
default_grafana_config['dashboard'] = default_dashboard
default_grafana_config['datasource'] = default_datasource

def edit_grafana(
        action : list = [''],
        debug : bool = False,
        **kw
    ):
    """
    Open Grafana configuration files for editing
    """

    from meerschaum.config._edit import general_edit_config
    from meerschaum.config._paths import GRAFANA_DATASOURCE_PATH, GRAFANA_DASHBOARD_PATH
    files = {
        'data' : GRAFANA_DATASOURCE_PATH,
        'datasource' : GRAFANA_DATASOURCE_PATH,
        'datasources' : GRAFANA_DATASOURCE_PATH,
        'datasource.yaml' : GRAFANA_DATASOURCE_PATH,
        'datasources.yaml' : GRAFANA_DATASOURCE_PATH,
        'dash' : GRAFANA_DASHBOARD_PATH,
        'dashboard' : GRAFANA_DASHBOARD_PATH,
        'dashboard.yaml' : GRAFANA_DASHBOARD_PATH,
    }
    return general_edit_config(action=action, files=files, default='datasource', debug=debug)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Expose plugin management APIs from the `meerschaum.plugins` module.
"""

from __future__ import annotations
from meerschaum.utils.typing import Callable, Any, Union, Optional, Dict, List
from meerschaum.actions.arguments._parser import add_plugin_argument

_api_plugins : dict = {}

def make_action(
        function : Callable[[Any], Any],
        shell : bool = False,
        debug : bool = False
    ) -> Callable[[Any], Any]:
    """
    Make a function a Meerschaum action. Useful for plugins that are adding multiple actions.

    Usage:
    ```
    >>> from meerschaum.plugins import make_action
    >>> 
    >>> @make_action
    ... def my_action(**kw):
    ...     print('foo')
    ...     return True, "Success"
    >>> 
    ```
    """
    from meerschaum.actions import __all__ as _all, actions
    from meerschaum.utils.formatting import pprint
    from meerschaum.utils.misc import add_method_to_class

    if debug:
        from meerschaum.utils.debug import dprint
        dprint(
            f"Adding action '{function.__name__}' from plugin " +
            f"'{function.__module__.split('.')[-1]}'..."
        )

    if function.__name__ not in _all:
        _all.append(function.__name__)
    actions[function.__name__] = function
    return function

def api_plugin(function) -> Callable[[Any], Any]:
    """
    Execute function when initializing the Meerschaum API module.
    Useful for lazy-loading heavy plugins only when the API is started,
    such as when editing the `meerschaum.api.app` FastAPI app.

    The FastAPI app will be passed as the only parameter.

    Usage:
    ```
    >>> from meerschaum.plugins import api_plugin
    >>> 
    >>> @api_plugin
    >>> def initialize_plugin(app):
    ...     @app.get('/my/new/path')
    ...     def new_path():
    ...         return {'message' : 'It works!'}
    >>> 
    ```
    """
    global _api_plugins
    if function.__module__ not in _api_plugins:
        _api_plugins[function.__module__] = []
    _api_plugins[function.__module__].append(function)
    return function

def import_plugins(plugins_to_import: Union[str, List[str], None] = None) -> Union[
        'ModuleType', Tuple['ModuleType', None]
    ]:
    """
    Import the Meerschaum plugins directory.

    :param plugins_to_import:
        If provided, only import the specified plugins.
        Otherwise import the entire plugins module. May be a string, list, or `None`.
        Defaults to `None`.
    """
    global __path__
    import sys
    from meerschaum.config._paths import (
        PLUGINS_RESOURCES_PATH, PLUGINS_ARCHIVES_RESOURCES_PATH, PLUGINS_INIT_PATH
    )
    PLUGINS_RESOURCES_PATH.mkdir(parents=True, exist_ok=True)
    PLUGINS_INIT_PATH.touch()

    if isinstance(plugins_to_import, str):
        plugins_to_import = [plugins_to_import]

    from meerschaum.utils.warnings import error, warn
    if str(PLUGINS_RESOURCES_PATH.parent) not in sys.path:
        sys.path.insert(0, str(PLUGINS_RESOURCES_PATH.parent))
    if str(PLUGINS_RESOURCES_PATH.parent) not in __path__:
        __path__.append(str(PLUGINS_RESOURCES_PATH.parent))

    if not plugins_to_import:
        try:
            import plugins
        except ImportError:
            plugins = None
    else:
        from meerschaum.utils.packages import attempt_import
        plugins = attempt_import(
            *[('plugins.' + p) for p in plugins_to_import],
            install=False, warn=False, lazy=False, venv=None,
        )

    if plugins is None:
        warn(f"Failed to import plugins.", stacklevel=3)

    if str(PLUGINS_RESOURCES_PATH.parent) in sys.path:
        sys.path.remove(str(PLUGINS_RESOURCES_PATH.parent))

    return plugins

def load_plugins(debug : bool = False, shell : bool = False) -> None:
    """
    Import Meerschaum plugins and update the actions dictionary.
    """
    ### append the plugins modules
    from inspect import isfunction, getmembers
    from meerschaum.actions import __all__ as _all, modules, make_action
    from meerschaum.utils.packages import get_modules_from_package
    if debug:
        from meerschaum.utils.debug import dprint

    _plugins_names, plugins_modules = get_modules_from_package(
        import_plugins(),
        names = True,
        recursive = True,
        modules_venvs = True
    )
    _all += _plugins_names
    modules += plugins_modules
    for module in plugins_modules:
        for name, func in getmembers(module):
            if not isfunction(func):
                continue
            if name == module.__name__.split('.')[-1]:
                make_action(func, **{'shell' : shell, 'debug' : debug})

def reload_plugins(plugins : Optional[List[str]] = None, debug : bool = False) -> None:
    """
    Reload plugins back into memory.
    """
    import sys
    if debug:
        from meerschaum.utils.debug import dprint

    if not plugins:
        plugins = get_plugins_names()
    for plugin_name in plugins:
        if debug:
            dprint(f"Reloading plugin '{plugin_name}'...")
        mod_name = 'plugins.' + str(plugin_name)
        if mod_name in sys.modules:
            del sys.modules[mod_name]
    load_plugins(debug=debug)

def get_plugins_names() -> Optional[List[str]]:
    from meerschaum.utils.packages import get_modules_from_package
    from meerschaum.utils.warnings import warn, error
    try:
        names = get_modules_from_package(import_plugins(), names=True, recursive=True)[0]
    except Exception as e:
        names = None
        warn(e, stacklevel=3)
    return names

def get_plugins_modules() -> Optional[List['ModuleType']]:
    from meerschaum.utils.packages import get_modules_from_package
    from meerschaum.utils.warnings import warn, error
    try:
        modules = get_modules_from_package(import_plugins(), recursive=True)
    except Exception as e:
        modules = None
        warn(e, stacklevel=3)
    return modules

def get_data_plugins() -> List['ModuleType']:
    """
    Return plugins which contain `fetch()` or `sync()` functions.
    """
    import inspect
    mods = get_plugins_modules()
    data_names = {'sync', 'fetch'}
    data_plugins = []
    for m in mods:
        for name, ob in inspect.getmembers(m):
            if not inspect.isfunction(ob):
                continue
            if name not in data_names:
                continue
            data_plugins.append(m)
    return data_plugins



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Delete a Pipe's contents and registration
"""

def delete(
        self,
        debug : bool = False,
        **kw
    ) -> tuple:
    """
    Call the Pipe's instance connector's delete_pipe method
    """
    result = self.instance_connector.delete_pipe(self, debug=debug, **kw)
    if not isinstance(result, tuple):
        return False, f"Received unexpected result from '{self.instance_connector}': {result}"
    if result[0]:
        to_delete = ['_id', '_attributes', '_parameters', '_columns']
        for member in to_delete:
            if member in self.__dict__:
                del self.__dict__[member]
    return result



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Register Pipes via the Meerschaum API
"""

from __future__ import annotations
from meerschaum.utils.typing import Any, Optional, Dict

from meerschaum.api import (
    fastapi,
    app,
    endpoints,
    get_api_connector,
    pipes,
    get_pipe,
    _get_pipes,
    manager,
    debug
)
from meerschaum.api.models import MetaPipe
from meerschaum.utils.packages import attempt_import
from meerschaum.utils.misc import is_pipe_registered, round_time
import meerschaum._internal.User
import datetime
pipes_endpoint = endpoints['pipes']

@app.post(pipes_endpoint)
def register_pipe(
        pipe : MetaPipe,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ):
    """
    Register a new pipe.
    """
    from meerschaum.config import get_config
    allow_pipes = get_config('system', 'api', 'permissions', 'registration', 'pipes', patch=True)
    if not allow_pipes:
        return False, (
            "The administrator for this server has not allowed pipe registration.\n\n" +
            "Please contact the system administrator, or if you are running this server, " +
            "open the configuration file with `edit config system` and search for 'permissions'." +
            " Under the keys `api:permissions:registration`, " +
            "you can toggle various registration types."
        )
    pipe_object = get_pipe(pipe.connector_keys, pipe.metric_key, pipe.location_key)
    if is_pipe_registered(pipe_object, pipes(refresh=True)):
        raise fastapi.HTTPException(status_code=409, detail="Pipe already registered")
    pipe_object.parameters = pipe.parameters
    results = get_api_connector().register_pipe(pipe_object)
    pipes(refresh=True)

    return results

@app.patch(pipes_endpoint)
def edit_pipe(
        pipe : MetaPipe,
        patch : bool = False,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ):
    """
    Edit a Pipe's parameters.
    patch : bool : False
        If patch is True, update the existing parameters by cascading.
        Otherwise overwrite the parameters (default)
    """
    from meerschaum.utils.debug import dprint
    pipe_object = get_pipe(pipe.connector_keys, pipe.metric_key, pipe.location_key)
    pipe_object.parameters = pipe.parameters
    if not is_pipe_registered(pipe_object, pipes(refresh=True)):
        raise fastapi.HTTPException(status_code=404, detail="Pipe is not registered.")

    results = get_api_connector().edit_pipe(pipe=pipe_object, patch=patch)

    pipes(refresh=True)
    return results

@app.get(pipes_endpoint + '/keys')
async def fetch_pipes_keys(
        connector_keys : str = "[]",
        metric_keys : str = "[]",
        location_keys : str = "[]",
        params : str = "{}",
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
        debug : bool = False
    ) -> list:
    """
    Get a list of tuples of all registered Pipes' keys.
    """
    import json

    keys = get_api_connector().fetch_pipes_keys(
        connector_keys = json.loads(connector_keys),
        metric_keys = json.loads(metric_keys),
        location_keys = json.loads(location_keys),
        params = json.loads(params),
        debug = debug
    )
    return keys

@app.get(pipes_endpoint)
async def get_pipes(
        connector_keys : str = "",
        metric_keys : str = "",
        location_keys : str = "",
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
        debug : bool = False
    ) -> dict:
    """
    Get all registered Pipes with metadata, excluding parameters.
    """
    from meerschaum.utils.misc import replace_pipes_in_dict
    kw = {'debug' : debug, 'mrsm_instance' : get_api_connector()}
    if connector_keys != "":
        kw['connector_keys'] = connector_keys
    if metric_keys != "":
        kw['metric_keys'] = metric_keys
    if location_keys != "":
        kw['location_keys'] = location_keys
    return replace_pipes_in_dict(_get_pipes(**kw), str)

@app.get(pipes_endpoint + '/{connector_keys}')
async def get_pipes_by_connector(
        connector_keys : str,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> dict:
    """
    Get all registered Pipes by connector_keys with metadata, excluding parameters.
    """
    from meerschaum.utils.misc import replace_pipes_in_dict
    if connector_keys not in pipes():
        raise fastapi.HTTPException(status_code=404, detail=f"connector_keys '{connector_keys}' not found.")
    return replace_pipes_in_dict(pipes()[connector_keys], str)

@app.get(pipes_endpoint + '/{connector_keys}/{metric_key}')
async def get_pipes_by_connector_and_metric(
        connector_keys : str,
        metric_key : str,
        parent : bool = False,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ):
    """
    Get all registered Pipes by connector_keys and metric_key with metadata, excluding parameters.

    parent : bool (default False)
        Return the parent Pipe (location_key is None)
    """
    from meerschaum.utils.misc import replace_pipes_in_dict
    if connector_keys not in pipes():
        raise fastapi.HTTPException(status_code=404, detail=f"connector_keys '{connector_keys}' not found.")
    if metric_key not in pipes()[connector_keys]:
        raise fastapi.HTTPException(status_code=404, detail=f"metric_key '{metric_key}' not found.")
    if parent:
        return pipes()[connector_keys][metric_key][None]
    return replace_pipes_in_dict(pipes()[connector_keys][metric_key], str)

@app.get(pipes_endpoint + '/{connector_keys}/{metric_key}/{location_key}')
async def get_pipes_by_connector_and_metric_and_location(
        connector_keys : str,
        metric_key : str,
        location_key : str,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ):
    """
    Get a specific Pipe with metadata, excluding parameters.
    """
    if connector_keys not in pipes():
        raise fastapi.HTTPException(status_code=404, detail=f"connector_keys '{connector_keys}' not found.")
    if metric_key not in pipes()[connector_keys]:
        raise fastapi.HTTPException(status_code=404, detail=f"metric_key '{metric_key}' not found.")
    if location_key in ('[None]', 'None', 'null'):
        location_key = None
    if location_key not in pipes()[connector_keys][metric_key]:
        raise fastapi.HTTPException(status_code=404, detail=f"location_key '{location_key}' not found.")

    return str(pipes()[connector_keys][metric_key][location_key])

@app.get(pipes_endpoint + '/{connector_keys}/{metric_key}/{location_key}/sync_time')
def get_sync_time(
        connector_keys : str,
        metric_key : str,
        location_key : str,
        params : dict = None,
        debug : bool = False,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> 'datetime.datetime':
    """
    Get a Pipe's latest datetime value.
    """
    if location_key == '[None]':
        location_key = None
    pipe = get_pipe(connector_keys, metric_key, location_key)
    if is_pipe_registered(pipe, pipes()):
        return pipe.get_sync_time(params=params, debug=debug)

@app.post(pipes_endpoint + '/{connector_keys}/{metric_key}/{location_key}/data')
def sync_pipe(
        connector_keys : str,
        metric_key : str,
        location_key : str,
        data : dict = {},
        check_existing : bool = True,
        blocking : bool = True,
        force : bool = False,
        workers : Optional[int] = None,
        columns : Optional[str] = None,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
        debug : bool = False,
    ) -> tuple:
    """
    Add data to an existing Pipe.
    """
    from meerschaum.utils.misc import parse_df_datetimes
    from meerschaum import Pipe
    import json
    p = get_pipe(connector_keys, metric_key, location_key)
    if not p.columns and columns is not None:
        p.columns = json.loads(columns)
    if not p.columns and not is_pipe_registered(p, pipes(refresh=True)):
        raise fastapi.HTTPException(
            status_code = 409,
            detail = "Pipe must be registered with the datetime column specified."
        )

    df = parse_df_datetimes(data)
    return p.sync(
        df,
        debug = debug,
        check_existing = check_existing,
        blocking = blocking,
        force = force,
        workers = workers
    )

@app.get(pipes_endpoint + '/{connector_keys}/{metric_key}/{location_key}/data')
def get_pipe_data(
        connector_keys : str,
        metric_key : str,
        location_key : str,
        begin : datetime.datetime = None,
        end : datetime.datetime = None,
        params : Optional[str] = None,
        orient : str = 'records',
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> str:
    """
    Get a Pipe's data. Optionally set query boundaries.
    """
    _params = {}
    if params is not None:
        import json
        try:
            _params = json.loads(params)
        except Exception as e:
            _params = None

    if not isinstance(_params, dict):
        raise fastapi.HTTPException(
            status_code = 409,
            detail = "Params must be a valid JSON-encoded dictionary.",
        )

    p = get_pipe(connector_keys, metric_key, location_key)
    if not is_pipe_registered(p, pipes(refresh=True)):
        raise fastapi.HTTPException(
            status_code = 409,
            detail = "Pipe must be registered with the datetime column specified"
        )

    #  chunks = p.get_data(
        #  begin = begin,
        #  end = end, 
        #  params = params,
        #  as_chunks = True,
        #  debug = debug
    #  )

    return fastapi.Response(
        content = p.get_data(
            begin = begin,
            end = end,
            params = _params, 
            debug = debug
        ).to_json(
            date_format = 'iso',
            orient = orient,
            date_unit = 'us',
        ),
        media_type = 'application/json',
        #  headers = {'chunk' : chunk, 'max_chunk' : max_chunk},
    )
@app.get(pipes_endpoint + '/{connector_keys}/{metric_key}/{location_key}/backtrack_data')
def get_backtrack_data(
        connector_keys : str,
        metric_key : str,
        location_key : str,
        begin : datetime.datetime = None,
        backtrack_minutes : int = 0,
        orient : str = 'records',
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> fastapi.Response:
    """
    Get a Pipe's data. Optionally set query boundaries
    """
    pipe = get_pipe(connector_keys, metric_key, location_key)
    if not pipe.get_columns('datetime', error=False):
        raise HTTPException(
            status_code = 400,
            detail = f"Cannot fetch backtrackdata for pipe '{pipe}' without a datetime column.",
        )
    df = pipe.get_backtrack_data(
            begin = begin,
            backtrack_minutes = backtrack_minutes,
            debug = debug
    )
    if df is None:
        return None
    js = df.to_json(
        date_format = 'iso',
        orient = orient,
        date_unit = 'us'
    )
    return fastapi.Response(
        content = js,
        media_type = 'application/json'
    )

@app.get(pipes_endpoint + '/{connector_keys}/{metric_key}/{location_key}/id')
def get_pipe_id(
        connector_keys : str,
        metric_key : str,
        location_key : str,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> int:
    """
    Get a Pipe's ID
    """
    try:
        pipe_id = int(
            get_pipe(
                connector_keys,
                metric_key,
                location_key
            ).id
        )
    except Exception as e:
        raise fastapi.HTTPException(status_code=404, detail=str(e))
    return pipe_id

@app.get(pipes_endpoint + '/{connector_keys}/{metric_key}/{location_key}/attributes')
def get_pipe_attributes(
        connector_keys : str,
        metric_key : str,
        location_key : str,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> dict:
    """
    Get a Pipe's attributes
    """
    return get_pipe(connector_keys, metric_key, location_key).attributes

@app.get(pipes_endpoint + '/{connector_keys}/{metric_key}/{location_key}/exists')
def get_pipe_exists(
        connector_keys : str,
        metric_key : str,
        location_key : str,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> dict:
    """
    Determine if a Pipe exists or not
    """
    return get_pipe(connector_keys, metric_key, location_key).exists()

@app.post(endpoints['metadata'])
def create_metadata(
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> bool:
    """
    Create Pipe metadata tables
    """
    from meerschaum.connectors.sql.tables import get_tables
    try:
        tables = get_tables(mrsm_instance=get_api_connector(), debug=debug)
    except Exception as e:
        raise fastapi.HTTPException(status_code=500, detail=str(e))
    return True

@app.get(pipes_endpoint + '/{connector_keys}/{metric_key}/{location_key}/rowcount')
def get_pipe_rowcount(
        connector_keys : str,
        metric_key : str,
        location_key : str,
        begin : Optional[datetime.datetime] = None,
        end : Optional[datetime.datetime] = None,
        params : Optional[Dict[str, Any]] = None,
        curr_user : 'meerschaum._internal.User.User' = fastapi.Depends(manager),
    ) -> int:
    """
    Return a Pipe's row count
    """
    return get_pipe(connector_keys, metric_key, location_key).get_rowcount(
        begin = begin,
        end = end,
        params = params,
        debug = debug
    )

@app.get(pipes_endpoint + '/{connector_keys}/{metric_key}/{location_key}/columns/types')
def get_pipe_columns_types(
        connector_keys : str,
        metric_key : str,
        location_key : str,
    ) -> dict:
    """
    Return a dictionary of a pipe's table's columns to their data types.

    E.g. An example dictionary for a small table.

    ```
    >>> {
    ...   'dt': 'TIMESTAMP WITHOUT TIMEZONE',
    ...   'id': 'BIGINT',
    ...   'val': 'DOUBLE PRECISION',
    ... }
    >>> 
    ```
    """
    return get_pipe(connector_keys, metric_key, location_key).get_columns_types(debug=debug)
